#! /usr/bin/env python3

"""Test qmi/tools/proc.py"""
import unittest
from unittest.mock import MagicMock, patch, call
import socket
import os, sys
import psutil
import subprocess
from argparse import Namespace, ArgumentError

import qmi
import qmi.tools.proc as proc

from qmi.core.config_defs import CfgQmi
from qmi.core.config_defs import CfgContext
from qmi.core.config_defs import CfgProcessManagement
from qmi.core.config_defs import CfgProcessHost
from qmi.core.config_struct import config_struct_from_dict
from qmi.core.exceptions import QMI_ApplicationException

from tests.patcher import PatcherQmiContext as QMI_Context
from tests.patcher import PatcherQmiRpcProxy as QMI_RpcProxy

CONFIG = {
    "ip": "10.10.10.10",  # Local IP address
    "port": 1032,
    "server_command": "test_config_server_command",
    "ssh_host": "test_config_ssh_host",
    "ssh_user": "test_config_ssh_user",
    "program_module": "test_program_module",
    "program_args": ["test_program_arg1", "--test program_arg2"],
}
CONFIG0 = {
    "ip": "172.16.4.2",
    "port": 5005,
    "program_module": "test_local_module",
    "program_args": ["test_program_arg0"],
}
CONTEXT_CFG = {
        "ContextName2": {"tcp_server_port": 1031},
        "ContextName1": {
            "host": CONFIG["ip"],
            "tcp_server_port": CONFIG["port"],
            "program_module": CONFIG["program_module"],
            "program_args": CONFIG["program_args"],
            "enabled": True,
        },
        "ContextName0": {
            "host": CONFIG0["ip"],
            "tcp_server_port": CONFIG0["port"],
            "program_module": CONFIG0["program_module"],
            "program_args": CONFIG0["program_args"],
            "connect_to_peers": ["ContextName1"],
            "enabled": True,
        }
    }


def _start_qmi_context():
    """Start qmi and initialize the context with a configuration. Returns the configuration."""
    qmi.start("ContextName2", context_cfg=CONTEXT_CFG)
    return {"config": CONFIG, "config0": CONFIG0}


def _build_mock_config(extra=False):
    context = QMI_Context
    qmi.context = context
    context.get_config = MagicMock(return_value=(config := MagicMock()))
    config.contexts = {
        "ContextName1": (ctxcfg1 := MagicMock()),
        "ContextName0": (ctxcfg0 := MagicMock()),
    }
    if extra:
        # Add an extra config
        config.contexts.update({"ContextName2": (ctxcfg2 := MagicMock())})
        ctxcfg2.host = "nonlocal"

    ctxcfg1.host = MagicMock()
    ctxcfg0.host = MagicMock()
    return context, config, ctxcfg1


class ProcessManagementClientTestCase(unittest.TestCase):

    def setUp(self):
        proc.subprocess = MagicMock(spec=subprocess)
        proc.subprocess.SubprocessError = subprocess.SubprocessError
        proc.open = MagicMock()
        self._config = _start_qmi_context()
        proc.print = MagicMock()
        proc._logger = MagicMock()

    def tearDown(self):
        qmi.stop()

    def test_pmc_init(self):
        """Test ProcessManagementClient.__init__, happy flow, configuration parameters."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()) as patcher:
            context, _, _ = _build_mock_config()
            context._config = CfgQmi(
                contexts={
                    "ContextName1": CfgContext(
                        host=self._config["config"]["ip"], tcp_server_port=0
                    ),
                },
                process_management=CfgProcessManagement(
                    hosts={
                        self._config["config"]["ip"]: CfgProcessHost(
                            server_command=self._config["config"]["server_command"],
                            ssh_host=self._config["config"]["ssh_host"],
                            ssh_user=self._config["config"]["ssh_user"],
                        )
                    }
                ),
            )
            context.get_config = MagicMock(return_value=context._config)
            proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
            proc.subprocess.Popen.assert_called_once_with(
                [
                    "ssh",
                    "-a",
                    "-x",
                    "-T",
                    "-q",
                    "-o",
                    "BatchMode yes",
                    "-l",
                    "test_config_ssh_user",
                    self._config["config"]["ssh_host"],
                    self._config["config"]["server_command"],
                ],
                stdin=proc.subprocess.PIPE,
                stdout=proc.subprocess.PIPE,
            )

        self.addCleanup(patcher)

    def test_pmc_init_default(self):
        """Test ProcessManagementClient.__init__, happy flow, default parameters."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()) as context:
            context().get_config = MagicMock(return_value=CfgQmi(
                contexts={
                    "ContextName1": CfgContext(
                        host=self._config["config"]["ip"], tcp_server_port=0
                    ),
                },
                process_management=CfgProcessManagement(
                    hosts={}
                ),
            ))
            proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
            proc.subprocess.Popen.assert_called_once_with(
                [
                    "ssh",
                    "-a",
                    "-x",
                    "-T",
                    "-q",
                    "-o",
                    "BatchMode yes",
                    self._config["config"]["ip"],
                    proc.DEFAULT_SERVER_COMMAND,
                ],
                stdin=proc.subprocess.PIPE,
                stdout=proc.subprocess.PIPE,
            )

        self.addCleanup(context)

    def test_pmc_init_os_error(self):
        """Test ProcessManagementClient.__init__, os error on Popen."""
        with patch("qmi.tools.proc.subprocess.Popen", MagicMock(side_effect=OSError)):
            with self.assertRaises(proc.ProcessException):
                proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")

    def test_pmc_close(self):
        """Test ProcessManagementClient.close, happy flow."""
        manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
        manager.close()
        manager._proc.stdin.close.assert_called_once_with()
        manager._proc.stdout.close.assert_called_once_with()
        manager._proc.wait.assert_called_once_with(timeout=2)

    def test_pmc_close_timeout(self):
        """Test whether ProcessManagementClient.close kills the process when a timeout happened."""
        manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
        with patch("qmi.tools.proc.qmi.context", MagicMock()), patch.object(
            manager._proc, "wait", side_effect=TimeoutError
        ):
            manager.close()

        manager._proc.kill.assert_called_once_with()

    def test_pmc_start_process(self):
        """Test ProcessManagementClient.start_process, happy flow."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
            with patch.object(
                manager._proc.stdout, "readline", return_value="OK 123".encode("ascii")
            ):
                rt_val = manager.start_process()
                manager._proc.stdin.write.assert_called_once_with(
                    b"START ContextName1\n"
                )
                manager._proc.stdin.flush.assert_called_once_with()
                self.assertEqual(123, rt_val)

    def test_pmc_start_process_value_error(self):
        """Test whether ProcessManagementClient.start_process value error pid."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
            with patch.object(
                manager._proc.stdout,
                "readline",
                return_value="OK INVALID".encode("ascii"),
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.start_process()

    def test_pmc_start_process_err(self):
        """Test whether ProcessManagementClient.start_process raises an exception when a error response is received."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
            with patch.object(
                manager._proc.stdout, "readline", return_value="ERR 123".encode("ascii")
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.start_process()

    def test_pmc_start_process_dead_server(self):
        """Test whether ProcessManagementClient.start_process raises an exception when no response is received."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")

            with patch.object(
                manager._proc.stdout, "readline", return_value="".encode("ascii")
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.start_process()

    def test_pmc_start_process_invalid_response(self):
        """Test whether ProcessManagementClient.start_process raises an exception when a invalid response is
        received."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")

            with patch.object(
                manager._proc.stdout,
                "readline",
                return_value="invalidresponse".encode("ascii"),
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.start_process()

    def test_pmc_stop_process(self):
        """Test ProcessManagementClient.stop_process, happy flow."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")

            with patch.object(
                manager._proc.stdout, "readline", return_value="OK 1".encode("ascii")
            ):
                rt_val = manager.stop_process(123)
                manager._proc.stdin.write.assert_called_once_with(
                    b"STOP ContextName1 123\n"
                )
                manager._proc.stdin.flush.assert_called_once_with()
                self.assertTrue(rt_val)

    def test_pmc_stop_process_value_error(self):
        """Test whether ProcessManagementClient.stop_process value error pid."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")
            with patch.object(
                manager._proc.stdout,
                "readline",
                return_value="OK INVALID".encode("ascii"),
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.stop_process(123)

    def test_pmc_stop_process_dead_server(self):
        """Test whether ProcessManagementClient.stop_process raises an exception when no response is returned."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")

            with patch.object(
                manager._proc.stdout, "readline", return_value="".encode("ascii")
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.stop_process(123)

    def test_pmc_stop_process_err(self):
        """Test whether ProcessManagementClient.stop_process raises an exception when an error is returned."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")

            with patch.object(
                manager._proc.stdout, "readline", return_value="ERR 1".encode("ascii")
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.stop_process(123)

    def test_pmc_stop_process_invalid_response(self):
        """Test whether ProcessManagementClient.stop_process raises an exception when a invalid response is returned
        by a context."""
        with patch("qmi.tools.proc.qmi.context", MagicMock()):
            _build_mock_config()
            manager = proc.ProcessManagementClient(self._config["config"]["ip"], "ContextName1")

            with patch.object(
                manager._proc.stdout,
                "readline",
                return_value="invalidresponse".encode("ascii"),
            ):
                with self.assertRaises(proc.ProcessException):
                    manager.stop_process(123)


class QmiProcMethodsTestCase(unittest.TestCase):

    def setUp(self):
        proc.subprocess = MagicMock(spec=subprocess)
        proc.subprocess.SubprocessError = subprocess.SubprocessError
        proc.open = MagicMock()
        proc.print = MagicMock()
        proc._logger = MagicMock()

    def test_is_local_host(self):
        """Test is_local_host, happy flow."""
        self.assertFalse(proc.is_local_host("0.0.0.0"))
        self.assertTrue(proc.is_local_host("127.0.0.0"))
        self.assertTrue(proc.is_local_host("::1"))
        self.assertFalse(proc.is_local_host("INVALID"))

        # Collect a local ip address that is not a loopback address.
        test_addr = None
        for addr in psutil.net_if_addrs().values():
            for a in addr:
                if a.family == socket.AF_INET and a.broadcast != None:
                    test_addr = a
                    break
            if test_addr:
                break

        if test_addr:
            self.assertTrue(proc.is_local_host(test_addr.address))

    @patch("qmi.tools.proc.socket")
    @patch("qmi.tools.proc.psutil", autospec=psutil)
    def test_is_local_host_assert_exceptions(self, psutil_patch, socket_patch):
        """Test is_local_host raises assertion exceptions."""
        getaddrinfo = unittest.mock.Mock(side_effect=[
            ([["a", "b", "c", "d", ["271.0.0.1"]]]),
            ([["a", "b", "c", "d", [127,0,0.1]]]),
            OSError("Test error")
        ])
        socket_patch.getaddrinfo = getaddrinfo
        net_if_addrs = unittest.mock.Mock(return_value={
            "some_if": ["foute_addr"]
        })
        psutil_patch.net_if_addrs = net_if_addrs
        psutil_patch._common.snicaddr = psutil._common.snicaddr
        # Except first the second assertion
        with self.assertRaises(AssertionError) as ass_err_2:
            proc.is_local_host("123.45.67.89")

        # Then except the first assertion
        with patch("qmi.tools.proc.str", spec_set=str, new=lambda x: 127) as str_patch:
            with self.assertRaises(AssertionError) as ass_err_1:
                proc.is_local_host("123.45.67.89")

        # As final step exception at getting address info returns False
        self.assertFalse(proc.is_local_host("123.45.67.89"))

        # Assert
        self.assertEqual("", str(ass_err_2.exception))
        self.assertEqual("AssertionError(\"'ip_address' unexpectedly not a string!\")", repr(ass_err_1.exception))

    def test_start_local_process(self):
        """Test start_local_process, happy flow."""
        with patch(
            "qmi.tools.proc.open", MagicMock(return_value=(fopen := MagicMock()))
        ), patch(
            "qmi.tools.proc.subprocess.Popen",
            MagicMock(return_value=(popen := MagicMock())),
        ), patch(
            "qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)
        ), patch(
            "sys.executable", (exec := MagicMock())
        ), patch(
            "os.environ.copy", MagicMock()
        ):
            context_name = "ContextName1"
            config = CfgQmi()
            for key in CONTEXT_CFG.keys():
                config.contexts.update({key: config_struct_from_dict(CONTEXT_CFG[key], CfgContext)})

            QMI_Context.get_config = MagicMock(return_value=config)
            ctxcfg = config.contexts[context_name]
            popen.pid = 0
            popen.poll = MagicMock(return_value=None)
            rt_val = proc.start_local_process(context_name)
            self.assertEqual(rt_val, popen.pid)
            proc.subprocess.Popen.assert_called_once_with(
                [exec, "-m", ctxcfg.program_module] + ctxcfg.program_args,
                stdin=proc.subprocess.DEVNULL,
                stdout=fopen,
                stderr=proc.subprocess.STDOUT,
                start_new_session=True,
                env=os.environ.copy(),
            )

    def test_start_local_process_winenv(self):
        """Test start_local_process, happy flow, windows environment."""
        mock_pid_parent = MagicMock()
        mock_pid_parent.children = MagicMock(return_value=[(mock_pid:=MagicMock())])
        with patch(
            "qmi.tools.proc.open", MagicMock(return_value=(fopen := MagicMock()))
        ), patch(
            "qmi.tools.proc.subprocess.Popen",
            MagicMock(return_value=(popen := MagicMock())),
        ), patch(
            "qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)
        ), patch(
            "sys.executable", (exec := MagicMock())
        ), patch(
            "os.environ.copy", MagicMock()
        ), patch("qmi.tools.proc.WINENV", True), patch('qmi.tools.proc.psutil.Process', MagicMock(return_value=mock_pid_parent)):
            _, _, ctxcfg = _build_mock_config()
            popen.poll = MagicMock(return_value=None)
            rt_val = proc.start_local_process("ContextName1")
            self.assertEqual(rt_val, mock_pid.pid)

    def test_start_local_process_stopped(self):
        """Test whether start_local_process raises an exception when the process is not successfully started."""
        with patch(
            "qmi.tools.proc.subprocess.Popen",
            MagicMock(return_value=(popen := MagicMock())),
        ), patch("qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)):
            popen.poll = MagicMock(
                return_value=True
            )  # returns something when the process is not running.
            with self.assertRaises(proc.ProcessException):
                proc.start_local_process("ContextName1")

    def test_start_local_process_not_exist(self):
        """Test start_local_process whether an exception is raised when the context doesn't exist in the
        configuration."""
        with self.assertRaises(proc.ProcessException):
            proc.start_local_process("SomeContextNameThatDoesntExistInConfiguration")

    def test_start_local_process_no_program_module(self):
        """Test start_local_process whether an exception is raised when no program module is configured."""
        _, _, ctxcfg = _build_mock_config()
        ctxcfg.program_module = None
        with self.assertRaises(proc.ProcessException):
            proc.start_local_process("ContextName1")

    def test_start_local_process_python_path(self):
        """Test whether start_local_process PYTHONPATH configuration sets the PYTHONPATH in the environment."""
        env_copy = {"PYTHONPATH": "SOMEPATH"}
        with patch(
            "qmi.tools.proc.subprocess.Popen",
            MagicMock(return_value=(popen := MagicMock())),
        ), patch("qmi.tools.proc.os", MagicMock()), patch(
            "qmi.tools.proc.os.environ.copy", MagicMock(return_value=env_copy)
        ), patch(
            "qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)
        ):
            _, _, ctxcfg = _build_mock_config()
            popen.poll = MagicMock(return_value=None)
            popen.pid = 0
            ctxcfg.python_path = "SOMEOTHERPATH"
            proc.start_local_process("ContextName1")
            self.assertEqual(env_copy["PYTHONPATH"], "SOMEOTHERPATH")

    def test_start_local_process_invalid_python_path(self):
        """Test whether start_local_process raises an exception when an invalid PYTHONPATH is not configured."""
        with patch("qmi.tools.proc.os.path.isdir", MagicMock(return_value=False)):
            _, _, _ = _build_mock_config()
            with self.assertRaises(proc.ProcessException):
                proc.start_local_process("ContextName1")

    def test_start_local_process_no_host(self):
        """Test whether start_local_process raises an exception when the host is not configured."""
        with patch("qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)):
            _, _, ctxcfg = _build_mock_config()
            ctxcfg.host = None
            with self.assertRaises(proc.ProcessException):
                proc.start_local_process("ContextName1")

    def test_start_local_process_output_file_with_cfg_output_dir(self):
        """Test whether start_local_process takes the configured output directory."""
        with patch(
            "qmi.tools.proc.subprocess.Popen",
            MagicMock(return_value=(popen := MagicMock())),
        ), patch(
            "qmi.tools.proc.time.strftime", MagicMock(return_value="strftime")
        ), patch(
            "qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)
        ):
            context, _, ctxcfg = _build_mock_config()
            ctxcfg.process_management.output_dir = "SomeOutputDir"
            popen.poll = MagicMock(return_value=None)
            popen.pid = 0
            context.resolve_file_name = MagicMock(return_value="ResolvedOutputDir")
            proc.start_local_process("ContextName1")
            proc.open.assert_called_once_with(
                os.path.join("ResolvedOutputDir", "ContextName1_strftime.out"), "a"
            )

    def test_start_local_process_output_file_open_oserror(self):
        """Test whether start_local_process raises an exception when the unable to open output folder."""
        with patch("qmi.tools.proc.open", MagicMock(side_effect=OSError)), patch(
            "qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)
        ):
            with self.assertRaises(proc.ProcessException):
                proc.start_local_process("ContextName1")

    def test_start_local_process_popen_oserror(self):
        """Test whether start_local_process raises an exception when the process raises an exception.."""
        with patch(
            "qmi.tools.proc.subprocess.Popen", MagicMock(side_effect=OSError)
        ), patch("qmi.tools.proc.os.path.isdir", MagicMock(return_value=True)):
            with self.assertRaises(proc.ProcessException):
                proc.start_local_process("ContextName1")

    def test_stop_local_process(self):
        """Test stop_local_process, happy flow."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(return_value=(process := MagicMock())),
        ):
            process.is_running = MagicMock(side_effect=[True, False])
            rt_val = proc.stop_local_process("ContextName1", 123)
            self.assertTrue(rt_val)
            proc.psutil.Process.assert_called_once_with(pid=123)

    def test_stop_local_process_no_process(self):
        """Test stop_local_process flow when no process is found."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(side_effect=psutil.NoSuchProcess(123)),
        ):
            rt_val = proc.stop_local_process("ContextName1", 123)
            self.assertFalse(rt_val)

    def test_stop_local_process_not_running(self):
        """Test stop_local_process flow when process is not running."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(return_value=(process := MagicMock())),
        ):
            process.is_running = MagicMock(return_value=False)
            rt_val = proc.stop_local_process("ContextName1", 123)
            self.assertFalse(rt_val)

    def test_stop_local_process_windows_permission_failure(self):
        """Test stop_local_process flow when unable to retrieve process status."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(return_value=(process := MagicMock())),
        ):
            process.is_running = MagicMock(side_effect=psutil.Error)
            with self.assertRaises(proc.ProcessException):
                proc.stop_local_process("ContextName1", 123)

    def test_stop_local_process_kill_no_process(self):
        """Test stop_local_process flow when kill occurs on a non running process."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(return_value=(process := MagicMock())),
        ):
            process.is_running = MagicMock(return_value=True)
            process.kill = MagicMock(side_effect=psutil.NoSuchProcess(123))
            rt_val = proc.stop_local_process("ContextName1", 123)
            self.assertFalse(rt_val)

    def test_stop_local_process_permision_error(self):
        """Test stop_local_process flow when kill results in an exception."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(return_value=(process := MagicMock())),
        ):
            process.is_running = MagicMock(return_value=True)
            process.kill = MagicMock(side_effect=psutil.Error)
            with self.assertRaises(proc.ProcessException):
                proc.stop_local_process("ContextName1", 123)

    def test_stop_local_process_not_killed(self):
        """Test stop_local_process flow unable to stop the process.."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(return_value=(process := MagicMock())),
        ):
            process.is_running = MagicMock(return_value=True)
            with self.assertRaises(proc.ProcessException):
                proc.stop_local_process("ContextName1", 123)

    def test_stop_local_process_wait_psutil_error(self):
        """Test stop_local_process flow when waiting results in an exception."""
        with patch(
            "qmi.tools.proc.psutil.Process",
            MagicMock(return_value=(process := MagicMock())),
        ):
            process.is_running = MagicMock(return_value=True)
            process.wait = MagicMock(side_effect=psutil.Error)
            with self.assertRaises(proc.ProcessException):
                proc.stop_local_process("ContextName1", 123)

    def _make_context_mock_peer(self):
        """Adapt the qmi context to use mocking for peer connections."""
        proxy = QMI_RpcProxy(QMI_Context(), None)
        qmi.context().make_peer_context_proxy = MagicMock(return_value=proxy)
        return proxy

    def test_start_process_localhost(self):
        """Test start_process localhost, happy flow."""
        with patch(
            "qmi.tools.proc.is_local_host", MagicMock(return_value=True)
        ), patch("qmi.tools.proc.start_local_process", MagicMock()):
            _, _, _ = _build_mock_config()
            rt_val = proc.start_process("ContextName1")
            self.assertEqual(rt_val, proc.start_local_process("ContextName1"))

    def test_start_process_remote(self):
        """Test start_process remotehost, happy flow."""
        with patch(
            "qmi.tools.proc.is_local_host", MagicMock(return_value=False)
        ), patch("qmi.tools.proc.ProcessManagementClient", MagicMock()):
            _, _, ctxcfg = _build_mock_config()
            rt_val = proc.start_process("ContextName1")
            self.assertEqual(
                rt_val,
                proc.ProcessManagementClient(
                    ctxcfg.host, "ContextName1"
                ).start_process(),
            )

    def test_start_no_ctxcfg(self):
        """Test start_process invalid configuration flow."""
        _, config, _ = _build_mock_config()
        config.contexts = {}
        with self.assertRaises(proc.ProcessException):
            proc.start_process("ContextName1")

    def test_start_no_host(self):
        """Test start_process invalid configuration flow."""
        _, _, ctxcfg = _build_mock_config()
        ctxcfg.host = None
        with self.assertRaises(proc.ProcessException):
            proc.start_process("ContextName1")

    def test_stop_process_localhost(self):
        """Test start_process localhost, happy flow."""
        with patch(
            "qmi.tools.proc.is_local_host", MagicMock(return_value=True)
        ), patch("qmi.tools.proc.stop_local_process", MagicMock()):
            _, _, _ = _build_mock_config()
            rt_val = proc.stop_process("ContextName1", 123)
            self.assertEqual(rt_val, proc.stop_local_process("ContextName1", 123))

    def test_stop_process_remote(self):
        """Test start_process remotehost, happy flow."""
        with patch(
            "qmi.tools.proc.is_local_host", MagicMock(return_value=False)
        ), patch("qmi.tools.proc.ProcessManagementClient", MagicMock()):
            _, _, ctxcfg = _build_mock_config()
            rt_val = proc.stop_process("ContextName1", 123)
            self.assertEqual(
                rt_val,
                proc.ProcessManagementClient(ctxcfg.host, "ContextName1").stop_process(
                    123
                ),
            )

    def test_stop_process_no_ctxcfg(self):
        """Test start_process invalid configuration flow."""
        _, config, _ = _build_mock_config()
        config.contexts = {}
        with self.assertRaises(proc.ProcessException):
            proc.stop_process("ContextName1", 123)

    def test_stop_process_no_host(self):
        """Test start_process invalid configuration flow."""
        _, _, ctxcfg = _build_mock_config()
        ctxcfg.host = None
        with self.assertRaises(proc.ProcessException):
            proc.stop_process("ContextName1", 123)

    def test_get_context_status(self):
        """Test get_context_status happy flow."""
        _build_mock_config()
        rt_val = proc.get_context_status("ContextName1")
        self.assertEqual(rt_val, (123, "SomeVersion"))

    def test_get_context_status_no_tcp(self):
        """Test get_context_status flow when no host is configured."""
        _, _, ctxcfg = _build_mock_config()
        ctxcfg.host = None
        with self.assertRaises(proc.ProcessException):
            proc.get_context_status("ContextName1")

    def test_get_context_status_os_error(self):
        """Test get_context_status flow when connecting results in an operation error."""
        with patch("qmi.tools.proc.qmi.context", QMI_Context) as context:
            context.get_config = MagicMock(return_value=(config := MagicMock()))
            config.contexts = {
                "ContextName1": (MagicMock()),
                "ContextName0": (MagicMock()),
            }
            context.connect_to_peer = MagicMock(side_effect=OSError)
            rt_val = proc.get_context_status("ContextName1")
            self.assertEqual(rt_val, (-1, ""))

        self.addCleanup(context)

    def test_get_context_status_connect_exception(self):
        """Test get_context_status flow when connecting results in an qmi error."""
        with patch("qmi.tools.proc.qmi.context", QMI_Context) as context:
            context.connect_to_peer = MagicMock(side_effect=proc.QMI_Exception)
            with self.assertRaises(proc.ProcessException):
                proc.get_context_status("ContextName1")

    def test_get_context_status_get_pid_exception(self):
        """Test get_context_status flow when unable to collect the PID."""
        proxy = QMI_RpcProxy
        proxy.get_pid = MagicMock(side_effect=proc.QMI_Exception)
        with patch("tests.patcher.PatcherQmiRpcProxy", proxy):
            with self.assertRaises(proc.ProcessException):
                proc.get_context_status("ContextName1")

        proxy.get_pid = lambda: 123

    def test_get_context_status_disconnect_exception(self):
        """Test get_context_status flow when disconnect results into an exception."""
        context, _, _ = _build_mock_config()
        context.connect_to_peer = MagicMock()
        context.disconnect_from_peer = MagicMock(
            side_effect=proc.QMI_UnknownNameException
        )
        proc.get_context_status("ContextName1")  # exception is 'passed'.

    def test_shutdown_context_hard(self):
        """Test shutdown_context, happy flow, hard shutdown."""
        proxy = QMI_RpcProxy
        future = MagicMock()
        future.wait = MagicMock()
        with patch("qmi.tools.proc.qmi.context", QMI_Context) as context, patch(
                "qmi.tools.proc.time", MagicMock()
        ), patch("qmi.tools.proc.CONTEXT_SHUTDOWN_TIMEOUT", 0.5):
            context.connect_to_peer.side_effect = None
            context.has_peer_context = MagicMock(side_effect=[True, False])
            context.make_peer_context_proxy = MagicMock(return_value=proxy)
            qmi.context = context
            proxy(context, None).rpc_nonblocking.shutdown_context = MagicMock(return_value=future)

            rt_val = proc.shutdown_context("ContextName1", cb := MagicMock())
            self.assertEqual(rt_val.responding, True)
            self.assertEqual(rt_val.pid, 123)
            self.assertEqual(rt_val.success, True)

            cb.assert_has_calls(
                [
                    call("soft shutdown"),
                    call(""),
                    call("hard shutdown"),
                    call(""),
                ]
            )

    def test_shutdown_context_soft(self):
        """Test shutdown_context, happy flow, soft shutdown."""
        with patch(
            "qmi.tools.proc.time", MagicMock()
        ), patch("qmi.tools.proc.CONTEXT_SHUTDOWN_TIMEOUT", 0.5):
            context, _, _ = _build_mock_config()
            proxy = self._make_context_mock_peer()
            future = MagicMock()
            future.wait = MagicMock()
            proxy.rpc_nonblocking.shutdown_context = MagicMock(return_value=future)
            context.has_peer_context = MagicMock(return_value=False)

            rt_val = proc.shutdown_context("ContextName1", cb := MagicMock())
            self.assertEqual(rt_val.responding, True)
            self.assertEqual(rt_val.pid, 123)
            self.assertEqual(rt_val.success, True)

            cb.assert_has_calls(
                [
                    call("soft shutdown"),
                    call(""),
                ]
            )

    def test_shutdown_context_no_host(self):
        """Test shutdown_context flow when no host is configured."""
        _, _, ctxcfg = _build_mock_config()
        ctxcfg.host = None
        with self.assertRaises(proc.ProcessException):
            proc.shutdown_context("ContextName1", MagicMock())

    def test_shutdown_context_connect_oserror(self):
        """Test shutdown_context flow when connecting results in an os error."""
        with patch("qmi.tools.proc.qmi.context", QMI_Context) as context:
            context.connect_to_peer = MagicMock(side_effect=OSError)
            rt_val = proc.shutdown_context("ContextName1", MagicMock())
            self.assertEqual(rt_val.responding, False)
            self.assertEqual(rt_val.pid, -1)
            self.assertEqual(rt_val.success, False)

        self.addCleanup(context)

    def test_shutdown_context_connect_exception(self):
        """Test shutdown_context flow when connecting results in a qmi error."""
        context, _, _ = _build_mock_config()
        self._make_context_mock_peer()
        context.connect_to_peer = MagicMock(side_effect=proc.QMI_Exception)
        with self.assertRaises(proc.ProcessException):
            proc.shutdown_context("ContextName1", MagicMock())

    def test_shutdown_context_pid_exception(self):
        """Test shutdown_context flow when get pid results in an exception."""
        with patch("qmi.tools.proc.qmi.context", QMI_Context) as context:
            context.has_peer_context = MagicMock(return_value=False)
            proxy = self._make_context_mock_peer()
            proxy.get_pid = MagicMock(side_effect=proc.QMI_Exception)
            with self.assertRaises(proc.ProcessException):
                proc.shutdown_context("ContextName1", MagicMock())

    def test_shutdown_context_soft_timeout(self):
        """Test shutdown_context flow when soft shutdown results in a timeout."""
        proxy = QMI_RpcProxy
        future = MagicMock()
        future.wait = MagicMock(side_effect=proc.QMI_RpcTimeoutException)
        with patch("qmi.tools.proc.qmi.context", QMI_Context) as context, patch(
            "qmi.tools.proc.time", MagicMock()
        ):
            proxy(context, None).rpc_nonblocking.shutdown_context = MagicMock(return_value=future)
            context.has_peer_context = MagicMock(return_value=False)

            proc.shutdown_context("ContextName1", cb := MagicMock())

            cb.assert_has_calls(
                [
                    call("soft shutdown"),
                    call("timeout"),
                    call(""),
                ]
            )

    def test_shutdown_context_soft_exception(self):
        """Test shutdown_context flow when soft shutdown results in a connection error."""
        proxy = self._make_context_mock_peer()
        future = MagicMock()
        future.wait = MagicMock(side_effect=proc.QMI_Exception)
        proxy.rpc_nonblocking.shutdown_context = MagicMock(return_value=future)
        with self.assertRaises(proc.ProcessException):
            proc.shutdown_context("ContextName1", MagicMock())

    def test_shutdown_context_responding_failure(self):
        """Test shutdown_context flow when context is not responding."""
        with patch("qmi.tools.proc.time", MagicMock()), patch(
                "qmi.tools.proc.CONTEXT_SHUTDOWN_TIMEOUT", 0.5
        ):
            context, _, _ = _build_mock_config()
            proxy = self._make_context_mock_peer()
            future = MagicMock()
            future.wait = MagicMock()
            proxy.rpc_nonblocking.shutdown_context = MagicMock(return_value=future)
            context.has_peer_context = MagicMock(side_effect=[True, True])

            rt_val = proc.shutdown_context("ContextName1", MagicMock())
            self.assertEqual(rt_val.responding, True)
            self.assertEqual(rt_val.pid, 123)
            self.assertEqual(rt_val.success, False)

    def test_shutdown_context_disconnect_exception(self):
        """Test shutdown_context flow when disconnect results in an error."""
        with patch("qmi.tools.proc.qmi.context", QMI_Context) as context, patch(
            "qmi.tools.proc.time", MagicMock()
        ), patch("qmi.tools.proc.CONTEXT_SHUTDOWN_TIMEOUT", 0.5):
            proxy = self._make_context_mock_peer()
            future = MagicMock()
            future.wait = MagicMock()
            proxy.rpc_nonblocking.shutdown_context = MagicMock(return_value=future)
            context.has_peer_context = MagicMock(side_effect=[True, False])
            context.disconnect_from_peer = MagicMock(
                side_effect=proc.QMI_UnknownNameException
            )

            proc.shutdown_context("ContextName1", MagicMock())  # exception is passed.

    def test_select_contexts(self):
        """Test select_contexts, happy flow."""
        _, config, _ = _build_mock_config()
        # Make the contexts "proper" contexts for the 'is_local_host' check.
        config.contexts["ContextName0"] = CfgContext()
        config.contexts["ContextName0"].host = "127.0.0.1"  # Local
        config.contexts["ContextName0"].enabled = True
        config.contexts["ContextName1"] = CfgContext()
        config.contexts["ContextName1"].host = "123.456.78.9"  # Not local
        config.contexts["ContextName1"].enabled = True
        # Select and assert
        contexts = proc.select_contexts(config)
        self.assertEqual(contexts[0], "ContextName1")
        # This patch we need for the 'is_local_host' check in the call.
        with patch(
                "qmi.tools.proc.get_context_status",
                MagicMock(
                    side_effect=[
                        (-1, "SomeVersion"),
                        (123, "SomeVersion"),
                    ]
                ),
        ):
            contexts = proc.select_local_contexts(config)
            self.assertEqual(contexts[0], "ContextName0")

    def test_select_contexts_not_in_config(self):
        """Test select_context_by_name when provided context is not configured."""
        with self.assertRaises(proc.QMI_ApplicationException):
            proc.select_context_by_name(MagicMock(), "InvalidContextName")

    def test_select_contexts_no_names(self):
        """Test select_contexts when no contexts are configured."""
        _, config, _ = _build_mock_config()
        config.contexts = {}
        with self.assertRaises(proc.QMI_ApplicationException):
            proc.select_contexts(config)

    def test_select_contexts_invalid_name(self):
        """Test select_context_by_name when an invalid context name is configured."""
        _, config, _ = _build_mock_config()
        config.contexts = {"!INVALID": MagicMock()}
        with self.assertRaises(proc.QMI_ApplicationException):
            proc.select_context_by_name(config, "!INVALID")

    def test_proc_server_start(self):
        """Test proc_server_start happy flow."""
        _, config, _ = _build_mock_config()
        with patch(
            "qmi.tools.proc.start_local_process", MagicMock(return_value=123)
        ), patch("qmi.tools.proc.is_local_host", MagicMock(return_value=True)), patch(
            "sys.stdout.flush", MagicMock()
        ), patch(
            "sys.stdin.readline", MagicMock(side_effect=["START ContextName1", ""])
        ):
            rt_val = proc.proc_server(config)
            proc.start_local_process.assert_called_once_with("ContextName1")
            self.assertEqual(rt_val, 0)

    def test_proc_server_start_no_ctxcfg(self):
        """Test proc_server_start when contexts are not configured."""
        with patch("sys.stdout.flush", MagicMock()), patch(
            "sys.stdin.readline", MagicMock(side_effect=["START ContextName1", ""])
        ), patch(
            "qmi.tools.proc.print", MagicMock()
        ):
            _, config, _ = _build_mock_config()
            config.contexts = {}
            proc.proc_server(config)
            proc.print.assert_called_with("ERR Unknown context")

    def test_proc_server_start_not_local_host(self):
        """Test proc_server_start when context is not on a local host."""
        with patch(
            "qmi.tools.proc.start_local_process", MagicMock(return_value=123)
        ), patch("qmi.tools.proc.is_local_host", MagicMock(return_value=False)), patch(
            "sys.stdout.flush", MagicMock()
        ), patch(
            "sys.stdin.readline", MagicMock(side_effect=["START ContextName1", ""])
        ), patch(
            "qmi.tools.proc.print", MagicMock()
        ):
            _, config, _ = _build_mock_config()
            proc.proc_server(config)
            proc.print.assert_called_with("ERR Context should not run on this host")
            # sadly the only valid way to test this.

    def test_proc_server_stop(self):
        """Test proc_server_stop, happy flow."""
        with patch(
            "qmi.tools.proc.stop_local_process", MagicMock(return_value=True)
        ), patch("qmi.tools.proc.is_local_host", MagicMock(return_value=True)), patch(
            "sys.stdout.flush", MagicMock()
        ), patch(
            "sys.stdin.readline", MagicMock(side_effect=["STOP ContextName1 123", ""])
        ), patch(
            "qmi.tools.proc.print", MagicMock()
        ):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_server(config)
            proc.stop_local_process.assert_called_once_with("ContextName1", 123)
            self.assertEqual(rt_val, 0)

    def test_proc_server_stop_invalid_pid(self):
        """Test proc_server_start when contexts are not configured."""
        with patch("sys.stdout.flush", MagicMock()), patch(
            "sys.stdin.readline", MagicMock(side_effect=["STOP ContextName1 0", ""])
        ):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_server(config)
            self.assertEqual(rt_val, 1)

    def test_proc_server_stop_no_ctxcfg(self):
        """Test proc_server_start when contexts are not configured."""
        with patch("sys.stdout.flush", MagicMock()), patch(
            "sys.stdin.readline", MagicMock(side_effect=["STOP ContextName1 123", ""])
        ), patch(
            "qmi.tools.proc.print", MagicMock()
        ):
            _, config, _ = _build_mock_config()
            config.contexts = {"ContextName1": None}
            proc.proc_server(config)
            proc.print.assert_called_with("ERR Unknown context")

    def test_proc_server_stop_no_local_host(self):
        """Test proc_server_stop when context is not localhost."""
        with patch(
            "qmi.tools.proc.stop_local_process", MagicMock(return_value=True)
        ), patch("qmi.tools.proc.is_local_host", MagicMock(return_value=False)), patch(
            "sys.stdout.flush", MagicMock()
        ), patch(
            "sys.stdin.readline", MagicMock(side_effect=["STOP ContextName1 123", ""])
        ), patch(
            "qmi.tools.proc.print", MagicMock()
        ):
            _, config, _ = _build_mock_config()
            proc.proc_server(config)
            proc.print.assert_called_with("ERR Context should not run on this host")

    def test_proc_server_value_error(self):
        """Test proc_server invalid input."""
        with patch("sys.stdout.flush", MagicMock()), patch(
            "sys.stdin.readline", MagicMock(side_effect="INVALID")
        ), patch("qmi.tools.proc.print", MagicMock()):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_server(config)
            self.assertEqual(rt_val, 1)

    def test_proc_start(self):
        """Test proc_start happy flow."""
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(
                side_effect=[
                    (-1, "SomeVersion"),
                    (123, "SomeVersion"),
                ]
            ),
        ), patch("qmi.tools.proc.start_process", MagicMock(return_value=123)):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_start(config, "ContextName1", False)
            self.assertEqual(rt_val, 0)

    def test_proc_start_all(self):
        """Test proc_start starts multiple contexts when 'context_name = None' and 'local = False'."""
        expected_calls = [
            call("ContextName1"),
            call("ContextName0"),
            call("ContextName2")
        ]
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(
                side_effect=[
                    (-1, "SomeVersion"),
                    (123, "SomeVersion"),
                    (-1, "SomeVersion"),
                    (123, "SomeVersion"),
                    (-1, "SomeVersion"),
                    (123, "SomeVersion"),
                ]
            ),
        ), patch("qmi.tools.proc.start_process", MagicMock(return_value=123)) as start_mock:
            _, config, _ = _build_mock_config(extra=True)
            rt_val = proc.proc_start(config, None, False)
            # Assert
            self.assertEqual(rt_val, 0)
            start_mock.assert_has_calls(expected_calls)

    def test_proc_start_local(self):
        """Test proc_start start only local contexts with 'context_name = None' and 'local = True'."""
        expected_calls = [
            call("ContextName1"),
            call("ContextName0")
        ]
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(
                side_effect=[
                    (-1, "SomeVersion"),
                    (123, "SomeVersion"),
                    (-1, "SomeVersion"),
                    (123, "SomeVersion"),
                ]
            ),
        ), patch("qmi.tools.proc.start_process", MagicMock(return_value=123)) as start_mock:
            _, config, _ = _build_mock_config()
            # Make sure 'ContextName0' and 'ContextName1' are locals
            config.contexts["ContextName1"].host = "127.0.0.1"
            config.contexts["ContextName0"].host = "127.0.0.1"
            rt_val = proc.proc_start(config, None, True)
            # Assert
            self.assertEqual(rt_val, 0)
            start_mock.assert_has_calls(expected_calls)

    def test_proc_start_already_running(self):
        """Test proc_start flow when invalid pid."""
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(
                return_value=(123, "SomeVersion"),
            ),
        ), patch("qmi.tools.proc.start_process", MagicMock(return_value=123)):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_start(config, "ContextName1", False)
            self.assertEqual(rt_val, 0)

    def test_proc_start_not_responding(self):
        """Test proc_start when context not responding."""
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(return_value=(-1, "SomeVersion")),
        ), patch("qmi.tools.proc.start_process", MagicMock(return_value=123)):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_start(config, "ContextName1", False)
            self.assertEqual(rt_val, 1)

    def test_proc_start_invalid_pid(self):
        """Test proc_start flow when invalid pid."""
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(
                side_effect=[
                    (-1, "SomeVersion"),
                    (456, "SomeVersion"),  # return a invalid pid
                ]
            ),
        ), patch("qmi.tools.proc.start_process", MagicMock(return_value=123)):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_start(config, "ContextName1", False)
            self.assertEqual(rt_val, 1)

    def test_proc_stop(self):
        """Test proc_stop happy flow."""
        _, config, _ = _build_mock_config()
        with patch(
            "qmi.tools.proc.shutdown_context",
            MagicMock(return_value=proc.ShutdownResult(True, 123, True)),
        ):
            rt_val = proc.proc_stop(config, "ContextName1", False)
            self.assertEqual(rt_val, 0)

    def test_proc_stop_all(self):
        """Test proc_stop stops all contexts. It should stop both local and non-local contexts in reverse order."""
        _, config, _ = _build_mock_config(extra=True)
        expected_order_shutdowns = [
            call("ContextName2", proc.show_progress_msg),
            call("ContextName0", proc.show_progress_msg),
            call("ContextName1", proc.show_progress_msg)
        ]
        with patch(
            "qmi.tools.proc.shutdown_context",
            MagicMock(return_value=proc.ShutdownResult(True, 123, True)),
        ) as shutdown_patch:
            config.contexts["ContextName1"].host = "127.0.0.1"  # Make one context local
            rt_val = proc.proc_stop(config, None, False)
            # Assert. Make sure the order is as expected for shutdown.
            self.assertEqual(rt_val, 0)
            shutdown_patch.assert_has_calls(expected_order_shutdowns, any_order=False)

    def test_proc_stop_locals(self):
        """Test proc_stop stopping all local contexts. Check that it is done in 'reverse' order. In _build_mock_config
        the contexts are added in order 'ContextName1', 'ContextName' and since Python 3.7 the dict order is guaranteed
        to be preserved as the input order, so we can test the reversion easily."""
        _, config, _ = _build_mock_config(extra=True)  # We have 3 contexts, but 'ContextName2' is not local
        expected_order_shutdowns = [
            call("ContextName0", proc.show_progress_msg), call("ContextName1", proc.show_progress_msg)
        ]
        with patch(
            "qmi.tools.proc.shutdown_context",
            MagicMock(return_value=proc.ShutdownResult(True, 123, True)),
        ) as shutdown_patch:
            # Make sure 'ContextName0' and 'ContextName1' are locals
            config.contexts["ContextName1"].host = "127.0.0.1"
            config.contexts["ContextName0"].host = "127.0.0.1"
            rt_val = proc.proc_stop(config, None, True)
            # Assert. Make sure the order is as expected for shutdown.
            self.assertEqual(rt_val, 0)
            shutdown_patch.assert_has_calls(expected_order_shutdowns, any_order=False)

    def test_proc_stop_not_responding(self):
        """Test proc_stop context not responding."""
        _, config, _ = _build_mock_config()
        with patch(
            "qmi.tools.proc.shutdown_context",
            MagicMock(return_value=proc.ShutdownResult(False, 123, False)),
        ):
            rt_val = proc.proc_stop(config, "ContextName1", False)
            self.assertEqual(rt_val, 0)

    def test_proc_stop_kill(self):
        """Test proc_stop kill has is killed."""
        _, config, _ = _build_mock_config()
        with patch(
            "qmi.tools.proc.shutdown_context",
            MagicMock(return_value=proc.ShutdownResult(True, 123, False)),
        ), patch("qmi.tools.proc.stop_process", MagicMock(return_value=True)):
            rt_val = proc.proc_stop(config, "ContextName1", False)
            self.assertEqual(rt_val, 0)

    def test_proc_stop_failure_responding(self):
        """Test proc_stop when context is not responding."""
        with patch(
            "qmi.tools.proc.shutdown_context",
            MagicMock(return_value=proc.ShutdownResult(True, 123, False)),
        ), patch("qmi.tools.proc.stop_process", MagicMock(return_value=False)):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_stop(config, "ContextName1", False)
            self.assertEqual(rt_val, 0)

    def test_proc_stop_exception(self):
        """Test proc_stop when shutdown resulted in exception."""
        with patch(
            "qmi.tools.proc.shutdown_context",
            MagicMock(side_effect=proc.ProcessException),
        ):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_stop(config, "ContextName1", False)
            self.assertEqual(rt_val, 1)

    def test_proc_status(self):
        """Test proc_status happy flow."""
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(return_value=(123, "SomeVersion")),
        ):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_status(config, "ContextName1")
            self.assertEqual(rt_val, 0)

    def test_proc_status_offline(self):
        """Test proc_status process offline."""
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(return_value=(-1, "SomeVersion")),
        ):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_status(config, "ContextName1")
            self.assertEqual(rt_val, 0)

    def test_proc_status_exception(self):
        """Test proc_status exception."""
        with patch(
            "qmi.tools.proc.get_context_status",
            MagicMock(side_effect=proc.ProcessException),
        ):
            _, config, _ = _build_mock_config()
            rt_val = proc.proc_status(config, "ContextName1")
            self.assertEqual(rt_val, 1)


class ArgParserTestCase(unittest.TestCase):

    def setUp(self) -> None:
        # suppress print-outs
        self._stderr = sys.stderr
        sys.stderr = open(os.devnull, 'w')  # redirect stderr to null device
        context_cfg = {
            "somectx": CfgContext(
                host="127.0.0.1",
                tcp_server_port=12345,
                program_module="test_program_module",
                enabled=True
            )
        }
        self.context_cfg = CfgQmi(contexts=context_cfg)
        ctx_patcher = patch("qmi.core.context.QMI_Context.get_config", return_value=self.context_cfg)
        self.addCleanup(ctx_patcher.stop)
        peer_patcher = patch("qmi.core.context.QMI_Context.connect_to_peer")
        self.addCleanup(peer_patcher.stop)
        ctx_patcher.start()
        peer_patcher.start()

    def tearDown(self) -> None:
        sys.stderr.close()
        sys.stderr = self._stderr

    @patch("builtins.print")
    def test_start_server_mode_fails(self, print_patch):
        """Test that server mode start fails with incompatible commands."""
        fail_all = Namespace(command="server", all=True, locals=False, context_name=None, config=None)
        fail_local = Namespace(command="server", all=False, locals=True, context_name=None, config=None)
        fail_context_name = Namespace(command="server", all=False, locals=False, context_name="somectx", config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=fail_all):
            retval = proc.main()
            self.assertEqual(retval, 1)

        print_patch.assert_called_with("ERROR: Can not specify server mode together with other options")
        print_patch.reset()

        with patch("argparse.ArgumentParser.parse_args", return_value=fail_local):
            retval = proc.main()
            self.assertEqual(retval, 1)

        print_patch.assert_called_with("ERROR: Can not specify server mode together with other options")
        print_patch.reset()

        with patch("argparse.ArgumentParser.parse_args", return_value=fail_context_name):
            retval = proc.main()
            self.assertEqual(retval, 1)

        print_patch.assert_called_with("ERROR: Can not specify server mode together with other options")

    @patch("builtins.print")
    @patch("sys.stderr", return_value=None)
    def test_start_all_fails(self, sys_patch, print_patch):
        """Test that starting all contexts fails with incompatible commands."""
        fail_context_name = Namespace(command="start", all=True, locals=False, context_name="fail_ctx", config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=fail_context_name):
            retval = proc.main()
            self.assertEqual(retval, 1)

        print_patch.assert_called_with("ERROR: Can not specify a context_name together with --all", file=sys_patch)
        print_patch.reset()

        # We cannot test the same way the mutex as the argparse goes to sys.exit() at ArgumentError. Patch!
        sys.argv = ["command", "start", "--all", "--locals"]
        with patch("qmi.tools.proc.argparse.ArgumentParser.error", side_effect=[ArgumentError(MagicMock(), "Fail!")]):
            with self.assertRaises(ArgumentError) as arg_err:
                retval = proc.main()

        self.assertEqual(retval, 1)
        self.assertEqual("argument : Fail!", str(arg_err.exception))

        # But one true and one false should not raise ArgumentError, even if it fails otherwise!
        sys.argv = ["command", "start", "--all"]
        retval = proc.main()
        self.assertEqual(retval, 1)

        sys.argv = ["command", "start", "--locals"]
        retval = proc.main()
        self.assertEqual(retval, 1)

    @patch("builtins.print")
    @patch("sys.stderr", return_value=None)
    def test_start_locals_fails(self, sys_patch, print_patch):
        """Test that starting local contexts fails with incompatible context name."""
        fail_context_name = Namespace(command="start", all=False, locals=True, context_name="fail_ctx", config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=fail_context_name):
            retval = proc.main()
            self.assertEqual(retval, 1)

        print_patch.assert_called_with("ERROR: Can not specify a context_name together with --locals", file=sys_patch)

    @patch("builtins.print")
    @patch("sys.stderr", return_value=None)
    def test_start_stop_restart_fails(self, sys_patch, print_patch):
        """Test that 'start', 'stop' and 'restart' commands require extra arguments."""
        commands = ["start", "stop", "restart"]
        for command in commands:
            fail_command = Namespace(command=command, all=False, locals=False, context_name=None, config=None)
            with patch("argparse.ArgumentParser.parse_args", return_value=fail_command):
                retval = proc.main()
                self.assertEqual(retval, 1)

            print_patch.assert_called_with("ERROR: Specify either: a context_name, or --all, or --local", file=sys_patch)
            print_patch.reset()

    def test_start_server_mode(self):
        """Test that server mode is started with command 'server'."""
        return_value = Namespace(command="server", all=False, locals=False, context_name=None, config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=return_value):
            with patch("sys.stdin.readline", return_value=None) as sys_readline:
                retval = proc.main()
                sys_readline.assert_called_once()
                self.assertEqual(retval, 0)

    def test_start_context_name(self):
        """Test that select_(local_)context is called with command 'start'."""
        ctx = self.context_cfg
        ctx_name = list(ctx.contexts.keys())[0]
        local_ctx = None
        start_ctx = Namespace(command="start", all=False, locals=False, context_name=ctx_name, config=None)
        start_local_ctx = Namespace(command="start", all=False, locals=True, context_name=local_ctx, config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=start_ctx),\
             patch("qmi.tools.proc.select_context_by_name", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_called_once_with(ctx, ctx_name)

        with patch("argparse.ArgumentParser.parse_args", return_value=start_local_ctx),\
             patch("qmi.tools.proc.select_local_contexts", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_called_once_with(ctx)

    def test_stop_context_name(self):
        """Test that select_(local_)context is called with command 'stop'."""
        ctx = self.context_cfg
        ctx_name = list(ctx.contexts.keys())[0]
        local_ctx = None
        stop_ctx = Namespace(command="stop", all=False, locals=False, context_name=ctx_name, config=None)
        stop_local_ctx = Namespace(command="stop", all=False, locals=True, context_name=local_ctx, config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=stop_ctx),\
             patch("qmi.tools.proc.select_context_by_name", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_called_once_with(ctx, ctx_name)

        with patch("argparse.ArgumentParser.parse_args", return_value=stop_local_ctx),\
             patch("qmi.tools.proc.select_local_contexts", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_called_once_with(ctx)

    def test_restart_context_name(self):
        """Test that select_(local_)context is called with command 'restart'.
        Actually this will call only `proc_stop` as we set it to except. But the logic is then tested.
        """
        ctx = self.context_cfg
        ctx_name = list(ctx.contexts.keys())[0]
        local_ctx = None
        restart_ctx = Namespace(command="restart", all=False, locals=False, context_name=ctx_name, config=None)
        restart_local_ctx = Namespace(command="restart", all=False, locals=True, context_name=local_ctx, config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=restart_ctx),\
             patch("qmi.tools.proc.select_context_by_name", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_called_once_with(ctx, ctx_name)

        with patch("argparse.ArgumentParser.parse_args", return_value=restart_local_ctx),\
             patch("qmi.tools.proc.select_local_contexts", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_called_once_with(ctx)

    def test_status_context_name(self):
        """Test that select_(local_)context is called with command 'restart'.
        Actually this will call only `proc_stop` as we set it to except. But the logic is then tested.
        """
        ctx = self.context_cfg
        ctx_name = list(ctx.contexts.keys())[0]
        status_ctx = Namespace(command="status", all=False, locals=False, context_name=ctx_name, config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=status_ctx),\
             patch("qmi.tools.proc.select_context_by_name", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_called_once_with(ctx, ctx_name)

    def test_unknown_command(self):
        """Test that unknown command just returns without trying to select a context.
        This test is not fully exclusive, but should demonstrate w.r.t. previous tests that this is the case.
        """
        unknown_ctx = Namespace(command="unknown", all=True, locals=False, context_name=None, config=None)
        unknown_local_ctx = Namespace(command="unknown", all=True, locals=True, context_name=None, config=None)
        with patch("argparse.ArgumentParser.parse_args", return_value=unknown_ctx),\
             patch("qmi.tools.proc.select_contexts", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_not_called()

        with patch("argparse.ArgumentParser.parse_args", return_value=unknown_local_ctx),\
             patch("qmi.tools.proc.select_local_contexts", side_effect=[QMI_ApplicationException("wrong")]) as sel_ctx:
            retval = proc.main()
            self.assertEqual(retval, 1)
            sel_ctx.assert_not_called()



import logging
import unittest
from unittest.mock import MagicMock

import qmi
import qmi.utils.context_managers
from qmi.core.exceptions import QMI_InstrumentException, QMI_InvalidOperationException
from qmi.core.transport import QMI_Transport
from qmi.instruments.instru_tech.instrutech_agc302 import InstruTech_AGC302


class MyTestCase(unittest.TestCase):

    def setUp(self) -> None:
        # Suppress logging.
        logging.getLogger("qmi.core.instrument").setLevel(logging.CRITICAL)
        self.transport = MagicMock(spec=QMI_Transport)
        qmi.start("AGC302_unit-test")
        self.instrument: InstruTech_AGC302 = qmi.make_instrument("pressure_gauge", InstruTech_AGC302, self.transport)

    def tearDown(self) -> None:
        qmi.stop()
        logging.getLogger("qmi.core.instrument").setLevel(logging.NOTSET)

    def test_read_gauge(self):
        # arrange
        self.transport.read_until.return_value = b'*   1.55E-12\r'

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            pressure = self.instrument.read_gauge()

        # assert
        self.transport.write.assert_called_once_with(b'#  RD\r')
        self.assertAlmostEqual(pressure, 1.55E-12)

    def test_read_gauge_sensor_off(self):
        # arrange
        self.transport.read_until.return_value = b'*   1.10E+03\r'

        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(QMI_InstrumentException):
                self.instrument.read_gauge()

    def test_read_gauge_sensor_not_exists(self):
        # arrange
        self.transport.read_until.return_value = b'*   9.90E+09\r'

        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(QMI_InstrumentException):
                self.instrument.read_gauge()

    def test_pressure_unit_torr(self):
        # arrange
        self.transport.read_until.return_value = b'*   TORR    \r'

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            unit = self.instrument.read_pressure_unit()

        # assert
        self.transport.write.assert_called_once_with(b'#  RU\r')
        self.assertEqual(unit, 'TORR')

    def test_set_pressure_unit_torr(self):
        # arrange
        self.transport.read_until.return_value = b'*   PROGM_OK\r'

        # act
        with qmi.utils.context_managers.open_close(self.instrument):
            self.instrument.set_pressure_unit('T')

        # assert
        self.transport.write.assert_called_once_with(b'#  SUT\r')

    def test_set_pressure_unsupported_unit(self):
        # arrange
        self.transport.read_until.return_value = b'*   PROGM_OK\r'

        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(ValueError):
                self.instrument.set_pressure_unit('A')

    def test_unopened_device(self):
        # arrange
        self.transport.read_until.return_value = b'*   PROGM_OK\r'

        # act and assert
        with self.assertRaises(QMI_InvalidOperationException):
            self.instrument.read_gauge()

    def test_syntax_error(self):
        # arrange
        self.transport.read_until.return_value = b'?   SYNTX_ER\r'

        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(QMI_InstrumentException):
                self.instrument.read_gauge()

        self.transport.write.assert_called_once_with(b'#  RD\r')

    def test_invalid_response(self):
        # arrange
        self.transport.read_until.return_value = b'sdfwe\r'

        # act and assert
        with qmi.utils.context_managers.open_close(self.instrument):
            with self.assertRaises(QMI_InstrumentException):
                self.instrument.read_gauge()

        self.transport.write.assert_called_once_with(b'#  RD\r')


if __name__ == '__main__':
    unittest.main()



"""Instrument driver for the Quantum Opus amp-sim module."""
import re

from qmi.instruments.stanford_research_systems.sim900 import Sim900
from qmi.core.instrument import QMI_Instrument
from qmi.core.rpc import rpc_method
from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_UsageException, QMI_InstrumentException


class AmpSimModule(QMI_Instrument):
    """Instrument driver for the Quantum Opus amp-sim module."""

    # Commands
    _GET_MODULE_COMMAND = "+A?"
    _GET_BIAS_COMMAND = "+B?"
    _SET_BIAS_COMMAND = "+B{};"
    _SET_ADC_LOW_GAIN_MODE_COMMAND = "+C{};"
    _GET_DEVICE_VOLTAGE = "+C?"
    _SET_RESET_EVENT_DURATION_COMMAND = "+D{};"
    _GET_RESET_EVENT_DURATION_COMMAND = "+D?"
    _SET_AUTO_RESET_ENABLE_COMMAND = "+E{};"
    _GET_AUTO_RESET_ENABLE_COMMAND = "+E?"
    _INITIATE_RESET_EVENT_COMMAND = "+F;"
    _INITIATE_AUTO_BIAS_COMMAND = "+G;"
    _STORE_BIAS_CURRENT_NVM_COMMAND = "+H{};"
    _GET_BIAS_CURRENT_NVM_COMMAND = "+H?;"
    _USE_BIAS_CURRENT_NVM_COMMAND = "+H;"

    # Regexes
    _FIX_GET_BIAS_NVM_RESPOSE_REGEX = r'^\d+'

    def __init__(self, context: QMI_Context, name: str, sim_900: Sim900, port: int) -> None:
        """Initialize driver.

        Arguments:
            name: Name for this instrument instance.
            sim_900: Instance of sim 900 module or RPC proxy of it, that hosts the amp-sim module.
            port: Port number in which the amp-sim module is located.
        """
        super().__init__(context, name)
        self._sim_900 = sim_900
        self._port = port

    @rpc_method
    def get_module_id(self) -> str:
        """Query module identification string.

        Returns:
            Module identification string.
        """
        return self._sim_900.ask_module(self._port, self._GET_MODULE_COMMAND)

    @rpc_method
    def get_device_bias_current(self) -> int:
        """Query device bias current.

        Query device bias in DAC units (0 ─ 65535) for 0 to 2.5 V bias through 100 kΩ bias resistor (0-25 μA bias
        current)

        Returns:
            Query device bias in DAC units.
        """
        response = self._sim_900.ask_module(self._port, self._GET_BIAS_COMMAND)
        return int(response)

    @rpc_method
    def set_device_bias_current(self, dac_units: int) -> None:
        """Set device bias current.

        Set device bias current in DAC units (d = 0 ─ 65535, integer values). 0 = off, 65535 = 25μA. The bias current
        defaults to zero upon power-up.

        Arguments:
            dac_units: Device bias in DAC units.
        """
        if not 0 <= dac_units <= 65535:
            msg = 'The device bias current should be between 0 and 65535 but is {}'.format(dac_units)
            raise QMI_UsageException(msg)
        command = self._SET_BIAS_COMMAND.format(dac_units)
        self._sim_900.send_terminated_message(self._port, command)

    @rpc_method
    def set_adc_low_gain_mode(self) -> None:
        """Set ADC gain to low_gain mode."""
        command = self._SET_ADC_LOW_GAIN_MODE_COMMAND.format(1)
        self._sim_900.send_terminated_message(self._port, command)

    @rpc_method
    def set_adc_high_gain_mode(self) -> None:
        """Set ADC gain to high_gain mode."""
        command = self._SET_ADC_LOW_GAIN_MODE_COMMAND.format(0)
        self._sim_900.send_terminated_message(self._port, command)

    @rpc_method
    def get_device_voltage(self) -> int:
        """Query device voltage.

        Query device voltage in ADC units. If module is in high-gain mode (setting 0) the voltage (in volts) is
        calculated as (ADC units)/65535*1.1. If the module is in low-gain mode (setting 1) the voltage (in volts)
        is calculated as (ADC units)/65535*5.0

        Returns:
            Query device bias in ADC units.
        """
        # NOTE: we cannot interpret this value unless we known if we're in high- or low-gain mode,
        #   and we have no way to query that setting. (Sigh.)
        response = self._sim_900.ask_module(self._port, self._GET_DEVICE_VOLTAGE)
        return int(response)

    @rpc_method
    def set_reset_event_duration(self, duration: int) -> None:
        """Set reset event duration.

        Set the Reset Event duration (d = 0 ─ 255, integer). Sets the length of time the device bias is set to zero
        when a latch condition (i.e., device in non-superconducting state) is detected. The duration is calculated
        in units of 10 ms.

        Arguments:
            duration: Duration in units of 10ms.
        """
        if not 0 <= duration <= 255:
            raise QMI_UsageException('The event duration should be between 0 and 255 but is {}'.format(duration))
        command = self._SET_RESET_EVENT_DURATION_COMMAND.format(duration)
        self._sim_900.send_terminated_message(self._port, command)

    @rpc_method
    def get_reset_event_duration(self) -> int:
        """Query Reset Event duration.

        Query the presently set Reset Event duration.The duration is represented in units of 10 ms.

        Returns:
            Duration in units of 10ms.
        """
        response = self._sim_900.ask_module(self._port, self._GET_RESET_EVENT_DURATION_COMMAND)
        return int(response)

    @rpc_method
    def set_auto_reset_enabled(self, enabled_flag: bool) -> None:
        """Enable or disable the auto-reset function.

        Enable or disable the auto-reset function. When enabled the module will monitor the voltage on the
        nanowire device and if it exceeds an internally set value indicating the device is no longer
        superconducting, an auto-reset event will be triggered. If disabled, a latch condition will be persistent
        until manually cleared by front-panel operation or by initiating a reset event through software.

        Arguments:
            enabled_flag: Boolean flag to enable or disable the auto-reset function.
        """
        enabled = int(bool(enabled_flag))
        command = self._SET_AUTO_RESET_ENABLE_COMMAND.format(enabled)
        self._sim_900.send_terminated_message(self._port, command)

    @rpc_method
    def get_auto_reset_enabled(self) -> bool:
        """Query the auto-reset function.

        Returns:
            Return value is True if enabled, False if disabled.
        """
        response = self._sim_900.ask_module(self._port, self._GET_AUTO_RESET_ENABLE_COMMAND)
        return bool(int(response))

    @rpc_method
    def initiate_reset_event(self) -> None:
        """Initiate a Reset Event.

        Initiate a Reset Event. The device bias is reduced to zero, held at zero for the Reset Event duration,
        and then returned to its previously set value.
        """
        self._sim_900.send_terminated_message(self._port, self._INITIATE_RESET_EVENT_COMMAND)

    @rpc_method
    def initiate_auto_bias_function(self) -> None:
        """Initiate the auto-bias function.

        Initiate the auto-bias function. The device bias current is swept up from zero until a latch condition
        is detected. This latching current is measured. A Reset Event is initiated and then the bias is increased back
        up to approximately 95% of the measured latching current. This function is most reliable when the incoming
        light on the device is minimized as incoming photons can cause a slightly lower measured latching current.
        Note: The auto-bias function can result in slightly different bias currents each time it is run.
        """
        self._sim_900.send_terminated_message(self._port, self._INITIATE_AUTO_BIAS_COMMAND)

    @rpc_method
    def store_bias_current_in_non_volatile_memory(self, dac_current: int) -> None:
        """Store a bias DAC value into non-volatile memory.

        Store a bias DAC value into non-volatile memory (d = 0 ─ 65535, integer values). 0=off, 65535=25μA. This value
        will be written into internal non-volatile memory for reproducible biasing of the device through the
        use_device_bias_from_non_volatile_memory.

        Note: The internal non-volatile memory is only guaranteed to survive for 100,000 write functions. This limit
        can be easily exceeded if an external program repeatedly calls this function.

        Arguments:
            dac_current: Bias current that needs to be stored in non-volatile memory in units of 25μA.
        """
        if not 0 <= dac_current <= 65535:
            msg = 'The event duration should be between 0 and 65535 but is {}'.format(dac_current)
            raise QMI_UsageException(msg)
        command = self._STORE_BIAS_CURRENT_NVM_COMMAND.format(dac_current)
        self._sim_900.send_terminated_message(self._port, command)

    @rpc_method
    def get_bias_current_from_non_volatile_memory(self) -> int:
        """Query DAC bias value stored in non-volatile memory.

        Return the value of the DAC bias stored in non-volatile memory. The bias value is stored in units of 25μA.

        Returns:
            Bias value stored in non-voletile memory in units of 25μA.
        """
        response = self._sim_900.ask_module(self._port, self._GET_BIAS_CURRENT_NVM_COMMAND)
        # Note: the device returns x.0\n instead of just a integer value (represented in chars). This is fixed by
        # extracting the first digits from the response string.
        match = re.match(self._FIX_GET_BIAS_NVM_RESPOSE_REGEX, response)
        if match is None:
            raise QMI_InstrumentException(
                "Expected response to start with digits but instead got: {}".format(response))
        first_digits = match.group(0)
        return int(first_digits)

    @rpc_method
    def use_device_bias_from_non_volatile_memory(self) -> None:
        """Sets the device bias to the value stored in non-volatile memory."""
        self._sim_900.send_terminated_message(self._port, self._USE_BIAS_CURRENT_NVM_COMMAND)



"""
Instrument driver for the Thorlabs TSP01 environmental sensor.
"""

import logging
from typing import List

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport, list_usbtmc_transports, UsbTmcTransportDescriptorParser

# Global variable holding the logger for this module.

_logger = logging.getLogger(__name__)


class Thorlabs_TSP01(QMI_Instrument):
    """Instrument driver for the Thorlabs TSP01 environmental sensor."""

    USB_VENDOR_ID = 0x1313
    USB_PRODUCT_ID = 0x80f8

    # Default response timeout in seconds.
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    @staticmethod
    def list_instruments() -> List[str]:
        """Return a list of QMI transport descriptors for attached power meters."""
        devices = []
        parser = UsbTmcTransportDescriptorParser
        for desc_str in list_usbtmc_transports():
            if parser.match_interface(desc_str):
                parameters = parser.parse_parameter_strings(desc_str)
                if parameters.get("vendorid") == Thorlabs_TSP01.USB_VENDOR_ID and \
                        parameters.get("productid") == Thorlabs_TSP01.USB_PRODUCT_ID:
                    devices.append(desc_str)
        return devices

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        """Initialize the instrument driver.

        Parameters:
            name: Name for this instrument instance.
            transport: Transport descriptor to access this instrument.
                For the Thorlabs TSP01, this is typically
                "usbtmc:vendorid=0x1313:productid=0x80f8:serialnr=<device_serial_number>"
        """
        super().__init__(context, name)
        self._transport = create_transport(transport)
        self._scpi_protocol = ScpiProtocol(self._transport,
                                           default_timeout=self.DEFAULT_RESPONSE_TIMEOUT)

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning (most) settings to their defaults."""
        self._check_is_open()
        self._scpi_protocol.write("*RST")
        self._scpi_protocol.ask("*OPC?")

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        response = self._scpi_protocol.ask("*IDN?")
        words = response.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException("Unexpected response to *IDN?, got {!r}".format(response))
        return QMI_InstrumentIdentification(vendor=words[0],
                                            model=words[1],
                                            serial=words[2],
                                            version=words[3])

    @rpc_method
    def get_errors(self) -> List[str]:
        """Read pending error messages from the instrument."""
        errors = []
        while True:
            response = self._scpi_protocol.ask("SYST:ERR?")
            if response.startswith("0,"):
                break
            errors.append(response)
        return errors

    @rpc_method
    def get_temperature(self, sensor: int) -> float:
        """Measure the temperature.

        Parameters:
            sensor: Select temperature sensor
                1 = internal sensor;
                2 = external thermistor 1;
                3 = external thermistor 2.
        Returns:
            Measured temperature in Celcius.
        """
        if sensor not in (1, 2, 3):
            raise QMI_InstrumentException("Unknown temperature sensor {}".format(sensor))
        cmd = "MEAS:TEMP{}?".format(sensor)
        response = self._scpi_protocol.ask(cmd)
        try:
            return float(response)
        except ValueError:
            raise QMI_InstrumentException("Unexpected response to {!r}, got {!r}".format(cmd, response))

    @rpc_method
    def get_humidity(self) -> float:
        """Measure the relative humidity (using the internal sensor)..

        Returns:
            Measured relative humidity in percent.
        """
        cmd = "MEAS:HUM?"
        response = self._scpi_protocol.ask(cmd)
        try:
            return float(response)
        except ValueError:
            raise QMI_InstrumentException("Unexpected response to {!r}, got {!r}".format(cmd, response))



import unittest
from unittest.mock import call, patch
import numpy as np

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException, QMI_UsageException
from qmi.core.transport import QMI_TcpTransport
from qmi.instruments.siglent import Siglent_Ssa3000x


class TestSSA3000X(unittest.TestCase):

    def setUp(self):
        # Add patches
        patcher = patch('qmi.instruments.siglent.ssa3000x.create_transport', spec=QMI_TcpTransport)
        self._transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        patcher = patch('qmi.instruments.siglent.ssa3000x.ScpiProtocol', autospec=True)
        self._scpi_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.instr = Siglent_Ssa3000x(QMI_Context("test_siglent"), name="siglent", transport_descr="")
        self.instr._TIMEOUT = 0.01  # Make testing faster
        self.instr.open()

    def tearDown(self):
        self.instr.close()

    def test_channel_exception(self):
        with self.assertRaises(QMI_UsageException):
            _ = self.instr.get_spectrum(0)
        with self.assertRaises(QMI_UsageException):
            _ = self.instr.get_spectrum(5)

    def test_get_id(self):
        """ Test case for `get_id(...)` function. """
        # arrange
        expected_id_string = "test_string"
        self._scpi_mock.ask.return_value = expected_id_string
        # act
        actual_val = self.instr.get_id()
        # assert
        self._scpi_mock.ask.assert_called_once_with("*IDN?", discard=True)
        self.assertEqual(actual_val, expected_id_string)

    def test_nullbyte(self):
        """ Test case for discarding null byte. """
        # arrange
        expected_id_string = "test_string"
        self._scpi_mock.ask.return_value = "\0" + expected_id_string
        # act
        actual_val = self.instr.get_id()
        # assert
        self._scpi_mock.ask.assert_called_once_with("*IDN?", discard=True)
        self.assertEqual(actual_val, expected_id_string)

    def test_get_freq_span(self):
        expected = 0.1
        self._scpi_mock.ask.return_value = str(expected)
        # act
        actual = self.instr.get_freq_span()
        # assert
        self._scpi_mock.ask.assert_called_once_with(":FREQ:SPAN?", discard=True)
        self.assertEqual(actual, expected)

    def test_get_freq_span_invalid(self):
        self._scpi_mock.ask.return_value = "unfloatable_string"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_freq_span()

    def test_set_freq_span(self):
        self.instr.set_freq_span(0)  # act
        self._scpi_mock.write.assert_called_with(":FREQ:SPAN 0.000000000 GHz")  # assert

        self.instr.set_freq_span(100)  # act
        self._scpi_mock.write.assert_called_with(":FREQ:SPAN 0.000000100 GHz")  # assert

        self.instr.set_freq_span(3.2e9)  # act
        self._scpi_mock.write.assert_called_with(":FREQ:SPAN 3.200000000 GHz")  # assert

    def test_set_freq_span_invalid(self):
        with self.assertRaises(ValueError):
            self.instr.set_freq_span(50)

    def test_get_freq_center(self):
        expected = 10
        self._scpi_mock.ask.return_value = str(expected)
        # act
        actual = self.instr.get_freq_center()
        # assert
        self._scpi_mock.ask.assert_called_once_with(":FREQ:CENT?", discard=True)
        self.assertEqual(actual, expected)

    def test_get_freq_center_invalid(self):
        self._scpi_mock.ask.return_value = "unfloatable_string"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_freq_center()

    def test_set_freq_center_with_zero_span(self):
        self._scpi_mock.ask.return_value = str(0)  # assume asking for span

        self.instr.set_freq_center(0)  # act
        self._scpi_mock.write.assert_called_with(":FREQ:CENT 0.000000000 GHz")  # assert

        self.instr.set_freq_center(3.2e9)  # act
        self._scpi_mock.write.assert_called_with(":FREQ:CENT 3.200000000 GHz")  # assert

    def test_set_freq_center(self):
        self._scpi_mock.ask.return_value = str(100)  # assume asking for span

        self.instr.set_freq_center(50)  # act
        self._scpi_mock.write.assert_called_with(":FREQ:CENT 0.000000050 GHz")  # assert

        self.instr.set_freq_center(3.199999950e9)  # act
        self._scpi_mock.write.assert_called_with(":FREQ:CENT 3.199999950 GHz")  # assert

        with self.assertRaises(ValueError):
            self.instr.set_freq_center(0)

        with self.assertRaises(ValueError):
            self.instr.set_freq_center(3.2e9)

    def test_get_freq_start(self):
        expected = 0.1
        self._scpi_mock.ask.return_value = str(expected)
        # act
        actual = self.instr.get_freq_start()
        # assert
        self._scpi_mock.ask.assert_called_once_with(":FREQ:STAR?", discard=True)
        self.assertEqual(actual, expected)

    def test_get_freq_start_invalid(self):
        self._scpi_mock.ask.return_value = "unfloatable_string"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_freq_start()

    def test_get_freq_stop(self):
        expected = 0.1
        self._scpi_mock.ask.return_value = str(expected)
        # act
        actual = self.instr.get_freq_stop()
        # assert
        self._scpi_mock.ask.assert_called_once_with(":FREQ:STOP?", discard=True)
        self.assertEqual(actual, expected)

    def test_get_freq_stop_invalid(self):
        self._scpi_mock.ask.return_value = "unfloatable_string"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_freq_stop()

    def test_get_trace_format(self):
        valid_formats = ["ASCii", "REAL"]
        for expected in valid_formats:
            with self.subTest(i=expected):
                # arrange
                self._scpi_mock.reset_mock()
                self._scpi_mock.ask.return_value = expected
                # act
                actual = self.instr.get_trace_format()
                # assert
                self._scpi_mock.ask.assert_called_once_with(":FORM?", discard=True)
                self.assertEqual(actual, expected)

    def test_get_trace(self):
        expected = [0.1, 0.2, 0.3]
        self._scpi_mock.ask.side_effect = ["ASCii", ",".join([str(val) for val in expected])+","]
        # act
        actual = self.instr.get_trace(1)
        # assert
        self._scpi_mock.ask.assert_has_calls([call(":TRAC:DATA? 1", discard=True)])
        self.assertListEqual(actual, expected)

    def test_get_trace_wrong_format(self):
        self._scpi_mock.ask.return_value = "REAL"
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_trace(1)

    def test_get_trace_wrong_data(self):
        self._scpi_mock.ask.side_effect = ["ASCii", "not_floatable_string,"]
        # act & assert
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_trace(1)

    def test_get_spectrum(self):
        expected = (np.array([1, 2, 3]), np.array([0.1, 0.2, 0.3]))
        self._scpi_mock.ask.side_effect = [
            str(expected[0][0]),
            str(expected[0][-1]),
            "ASCii",
            ",".join([str(val) for val in list(expected[1])])+","
        ]
        # act
        actual = self.instr.get_spectrum(1)
        # assert
        self.assertListEqual(list(actual[0]), list(expected[0]))
        self.assertListEqual(list(actual[1]), list(expected[1]))

    def test_get_spectrum_wrong_freq(self):
        self._scpi_mock.ask.side_effect = [
            "1.0",  # start
            "0.0",  # stop
            "ASCii",  # format
            "0.1,"  # trace data
        ]
        # act
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_spectrum(1)

    def test_get_spectrum_wrong_trace(self):
        self._scpi_mock.ask.side_effect = [
            "0.0",  # start
            "1.0",  # stop
            "ASCii",  # format
            ""  # trace data
        ]
        # act
        with self.assertRaises(QMI_InstrumentException):
            _ = self.instr.get_spectrum(1)



"""Driver for the NKT Photonics Interbus protocol.

The communication protocol to the laser driver is documented in the "Software Development Kit for NKT Photonics Instruments - Instruction Manual"
document, v 2.1.3, Oct 2018. [Reference 1].

Chapter 2 of this document describes the NKT Photonocs Interbus protocol (physical, framing, CRC, etc.)

This chapter says, at the start:

"This section describes the NKT Photonics Interbus protocol and module hardware integration.
 It is not to be confused with INTERBUS developed by Phoenix Contact."
"""

import enum
import logging
import time
from typing import NamedTuple, Optional

from qmi.core.transport import QMI_Transport
from qmi.core.exceptions import QMI_TimeoutException, QMI_InstrumentException


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class MessageType(enum.Enum):
    NACK              = 0 ; "Response. Message not understood, not applicable, or not allowed (Not acknowledged)."
    CRC_ERROR         = 1 ; "Response. CRC error in received message."
    BUSY              = 2 ; "Response. Cannot respond at the moment. Module too busy."
    ACK               = 3 ; "Response. Received message understood (Acknowledged)."
    READ              = 4 ; "Query. Read the contents of a register."
    WRITE             = 5 ; "Transmission. Write something to a register."
    SET_SINGLE_BIT    = 6 ; "Transmission. Write a logic one to a bit will set the corresponding bit in the register value. Logic zeros have no effect."
    CLEAR_SINGLE_BIT  = 7 ; "Transmission. Writing a logic one to a bit will clear the corresponding bit in the register value. Logic zeros have no effect."
    DATAGRAM          = 8 ; "Response. Register content returned; caused by a previous READ."
    TOGGLE_SINGLE_BIT = 9 ; "Transmission. Writing a logic one to a bit will invert (toggle) the corresponding bit in the register value. Logic zeros have no effect."


class RegisterNumber(enum.Enum):
    MODULE_TYPE    = 0x61 ; "Module type number. Is used to determine what type of module the host is communicating with."
    SERIAL_NUMBER  = 0x64 ; "Firmware version code."
    STATUS_BITS    = 0x66 ; "Status monitor for a module or system."
    ERROR_CODE     = 0x67 ; "Error code as a byte value."


def _crc_ccitt(crc:int , c: int) -> int:

    crc ^= (c << 8)

    for i in range(8):
        xorflag = ((crc & 0x8000) != 0)
        crc = (crc << 1) & 0xffff
        if xorflag:
            crc ^= 0x1021

    return crc


class InterbusMessage(NamedTuple):
    """Message for the NKT Interbus protocol."""
    destination: int
    source: int
    message_type: MessageType
    register_number: int
    data: bytes


def _encode_interbus_message(message: InterbusMessage) -> bytes:
    """Encode an InterbusMessage to a bytes buffer."""

    # First, verify parameters:
    #
    # Destination and source addresses:
    #
    # * Address 0 reserved for 'special purposes'.
    #
    # * Addresses 1.. 160 specifies an NKT module.
    #
    #   The following addresses are assigned be NKT and listed in Chapter 6 of [Reference 1]:
    #
    # * Addresses 161..255 are considered host addresses (from PC or other equipment)
    #
    # Message type:
    #
    # * See the MessageType enumeration type.

    ok = (1 <= message.destination <= 160)
    if not ok:
        raise ValueError("Bad 'destination' value: {!r}".format(message.destination))

    ok = (161 <= message.source <= 255)
    if not ok:
        raise ValueError("Bad 'source' value: {!r}".format(message.source))

    if message.data is not None:
        ok = (0 <= len(message.data) <= 240)
        if not ok:
            raise ValueError("Bad 'data' length: {!r}".format(len(message.data)))

    # Synthesize message.

    register_number = message.register_number

    if isinstance(register_number, RegisterNumber):
        register_number = register_number.value

    encoded_message = bytearray([message.destination, message.source, message.message_type.value, register_number])

    if message.data:
        encoded_message.extend(message.data)

    # Add 16-bit CRC
    crc = 0
    for b in encoded_message:
        crc = _crc_ccitt(crc, b)

    encoded_message.extend(bytes([crc // 256, crc % 256]))

    # Escape 3 specific byte values, making sure they don't occur in the actual message.
    # NOTE: it is important to start with 0x5e, otherwise values will be doubly substituted.
    for value in [0x5e, 0x0d, 0x0a]:
        encoded_message = encoded_message.replace(bytes([value]), bytes([0x5e, value + 0x40]))

    # Add Start-Of-Transmission (SOT) and End-Of-Transmission (EOT) markers.
    encoded_message_b = bytes([13]) + encoded_message + bytes([10])

    return encoded_message_b


def _decode_interbus_message(encoded_message: bytes) -> InterbusMessage:
    """Decode an InterbusMessage to a bytes buffer."""

    if len(encoded_message) < 8:
        raise ValueError("Invalid interbus message (too short)")

    # Check SOT and EOT
    ok = (encoded_message[0] == 13 and encoded_message[-1] == 10)
    if not ok:
        raise ValueError("Invalid interbus message (SOT/EOT)")

    encoded_message = encoded_message[1:-1]  # Get rid of SOT and EOT.

    # Unescape 3 values:
    for value in [0x0a, 0x0d, 0x5e]:
        encoded_message = encoded_message.replace(bytes([0x5e, value + 0x40]), bytes([value]))

    if len(encoded_message) < 6:
        raise ValueError("Invalid interbus message (too short)")

    # Check 16-bit CRC
    crc = 0
    for b in encoded_message:
        crc = _crc_ccitt(crc, b)

    if crc != 0:
        raise ValueError("Bad CRC while decoding NKT Photonics Interbus message.")

    encoded_message = encoded_message[:-2]

    message = InterbusMessage(encoded_message[0], encoded_message[1], MessageType(encoded_message[2]), encoded_message[3], encoded_message[4:])

    return message


class NKTPhotonicsInterbusProtocol:
    """This is a wrapper around a QMI_Transport that provides NKT Photonics Interbus communication primitives."""

    # According to NKT documentation, we must choose an address above 160.
    HOST_BASE_ADDRESS = 161

    # The NKT laser sometimes does not answer a request.
    # In that case we retry the request at most this number of times.
    MAX_RETRY_COUNT = 10

    def __init__(self, transport: QMI_Transport, timeout: float):
        self._transport = transport
        self._timeout = timeout
        self._source_toggle = 1

    def _read_message(self) -> InterbusMessage:
        """Read and decode an NKT Photonics Interbus message over the transport."""
        message = self._transport.read_until(message_terminator=b"\n", timeout=self._timeout)
        decoded_message = _decode_interbus_message(message)
        return decoded_message

    def _send_message(self, message: InterbusMessage) -> None:
        """Encode and send an NKT Photonics Interbus message over the transport."""
        encoded_message = _encode_interbus_message(message)
        self._transport.write(encoded_message)

    def _request_response(
            self,
            destination: int,
            message_type: MessageType,
            register_number: int,
            data: bytes
            ) -> InterbusMessage:
        """Attempt request/response sequence.

        Occasionally, the NKT fails to respond to a request.
        The cause of these failures is not known. The failure rate may vary;
        initial estimate was once per 100,000 requests; recent tests suggest
        one failure per 250 requests.

        When the NKT device fails to respond, we simply retry the request
        until it succeeds or until `MAX_RETRY_COUNT` is exceeded.

        Raises:
            QMI_TimeoutException: When timeout occurs while waiting for a response
                and maximum number of retries is reached.
            QMI_InstrumentException: When an invalid message is received
                and maximum number of retries is reached.
        """

        # Use a source address that is different from the previous request.
        # This avoids accidental processing of a stale response in case
        # the previous request was retried and got two responses.
        self._source_toggle = (self._source_toggle + 1) & 1
        source = self.HOST_BASE_ADDRESS + self._source_toggle

        request = InterbusMessage(destination, source, message_type, register_number, data)

        # Send query.
        self._send_message(request)

        # Wait for response.
        failure_counter = 0
        while True:

            try:
                response = self._read_message()
            except (ValueError, QMI_TimeoutException) as exc:
                # Catch timeout and badly formatted messages.
                _logger.warning("Error while reading response: %s", str(exc))

                # Check if we can try again. If not, re-raise the exception to be handled by the caller.
                failure_counter += 1
                if failure_counter > self.MAX_RETRY_COUNT:
                    if isinstance(exc, ValueError):
                        raise QMI_InstrumentException("Invalid interbus response") from exc
                    else:
                        raise

                # Resend the request
                self._send_message(request)

                # Wait again for a response.
                continue

            # We received a response. Check that it matches the request.
            if (response.source == request.destination) and (response.destination == request.source):
                break

            # We received a valid message that does not match our request.
            _logger.warning("Unexpected response src=%d,dst=%d while expecting src=%d,dst=%d",
                            response.source, response.destination, request.destination, request.source)

            # Check if we can try again. If not, raise an exception to be handled by the caller.
            failure_counter += 1
            if failure_counter > self.MAX_RETRY_COUNT:
                raise QMI_InstrumentException("Unexpected response message, maximum retry count reached")

            # Discard the mismatching message and wait for another message.
            # Do not resend the request, because a reply may already be pending.

        return response

    def get_register(self, destination: int, register_number: int) -> bytes:
        """Read the specified register."""

        response = self._request_response(destination, MessageType.READ, register_number, b"")

        if response.message_type == MessageType.NACK:
            raise QMI_InstrumentException(f"Got NACK response for get_register reg=0x{register_number:x}")
        if response.message_type != MessageType.DATAGRAM:
            raise QMI_InstrumentException(f"Unexpected response for get_register ({response})")
        if response.register_number != register_number:
            raise QMI_InstrumentException(f"Response does not match register number req.reg=0x{register_number:x}"
                                          + f" resp.reg=0x{response.register_number:x}")

        return response.data

    def set_register(self, destination: int, register_number: int, data: Optional[bytes]) -> None:
        """Write to the specified register."""

        if data is None:
            data = b""
        response = self._request_response(destination, MessageType.WRITE, register_number, data)

        if response.message_type == MessageType.NACK:
            raise QMI_InstrumentException(f"Got NACK response for set_register reg=0x{register_number:x}")
        if response.message_type != MessageType.ACK:
            raise QMI_InstrumentException(f"Unexpected response for set_register ({response})")
        if response.register_number != register_number:
            raise QMI_InstrumentException(f"Response does not match register number req.reg=0x{register_number:x}"
                                          + f" resp.reg=0x{response.register_number:x}")



import unittest, unittest.mock
import struct
import binascii
from typing import cast

from qmi.core.transport import QMI_SerialTransport
from qmi.instruments.thorlabs import Thorlabs_K10Cr1
from qmi.instruments.thorlabs.k10cr1 import _AptMsgHwGetInfo
import qmi.core.exceptions

# Number of microsteps per degree of rotation.
MICROSTEPS_PER_DEGREE = 409600.0 / 3.0
# Internal velocity setting for 1 degree/second.
VELOCITY_FACTOR = 7329109.0
# Internal accleration setting for 1 degree/second/second.
ACCELERATION_FACTOR = 1502.0


class TestThorlabsK10cr1(unittest.TestCase):
    def setUp(self):
        qmi.start("TestK10cr1OpenClose")
        self._transport_mock = unittest.mock.MagicMock(spec=QMI_SerialTransport)
        with unittest.mock.patch(
                'qmi.instruments.thorlabs.k10cr1.create_transport',
                return_value=self._transport_mock):
            self.instr: Thorlabs_K10Cr1 = qmi.make_instrument("instr", Thorlabs_K10Cr1, "transport_descriptor")
            self.instr = cast(Thorlabs_K10Cr1, self.instr)

    def tearDown(self):
        self._transport_mock.reset_mock()
        qmi.stop()

    def test_open_close(self):
        """Test opening and closing the instrument"""
        # We expect to write MESSAGE_ID 0x0005 (_AptMsgHwReqInfo)
        expected_write = struct.pack("<l", 0x0005) + b"P\x01"  # This is 5001 == 0x1389
        # We expect as response MESSAGE_ID 0x0006 (_AptMsgHwGetInfo)
        expected_read = struct.pack("<l", 0x0006)
        # The request+data has to be 90 bytes long and should include string "K10CR1" at right spot.
        self._transport_mock.read.return_value = expected_read + b"\x00" * 2 + b"CR1\0K10" * 12
        self.instr.open()
        self.instr.close()
        # Assert
        self._transport_mock.write.assert_called_with(expected_write)

    def test_open_close_expects_with_timeout_when_correct_message_type_not_received(self):
        """Test opening the instrument fails if we give wrong (but otherwise valid) message id to the call"""
        expected_exception = "Expected message type {} not received.".format(_AptMsgHwGetInfo)
        # We expect as response MESSAGE_ID 0x0006 (_AptMsgHwGetInfo) but we give 0x0412
        expected_read = struct.pack("<l", 0x0412) + b"\x00\x00"
        self._transport_mock.read.return_value = expected_read + b"K10CR1"
        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected_exception)

    def test_open_excepts_with_too_small_buffer(self):
        """Test opening the instrument time-outs"""
        # No response now to force exception
        self._transport_mock.read.return_value = b""
        with self.assertRaises(ValueError):
            self.instr.open()

    def test_open_excepts_with_timeout(self):
        """Test opening the instrument time-outs and raises QMI_InstrumentException"""
        # Long APT messages start with 0x80
        long_apt_msg = b"\x80" * 6
        expected_exception = "Received partial message (message_id=0x{:04x}, data_length={})".format(0x8080, 32896)
        # Make a long APT message to go into the loop where more data should be read, and cause timeout
        self._transport_mock.read.side_effect = [long_apt_msg, qmi.core.exceptions.QMI_TimeoutException]
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected_exception)

    def test_open_with_no_pending_message(self):
        """Test opening the instrument and no pending message in buffer."""
        # We expect to write MESSAGE_ID 0x0005 (_AptMsgHwReqInfo)
        expected_write = struct.pack("<l", 0x0005) + b"P\x01"  # This is 5001 == 0x1389
        # We expect as response MESSAGE_ID 0x0006 (_AptMsgHwGetInfo)
        expected_read = struct.pack("<l", 0x0006) + b"\x00" * 2 + b"CR1\0K10" * 12
        # The request+data has to be 90 bytes long and should include string "K10CR1" at right spot.
        self._transport_mock.read.side_effect = [qmi.core.exceptions.QMI_TimeoutException, expected_read]
        self.instr.open()
        self.instr.close()
        # Assert
        self._transport_mock.write.assert_called_with(expected_write)

    def test_open_excepts_with_unknown_message_id(self):
        """Test opening with unknown message ID fails"""
        expected_exception = "Received unknown message id 0x{:04x} from instrument".format(0x8080)
        # Make unexpected response MESSAGE_ID
        expected_read = struct.pack("l", 0x8080) + b"\x00\x00"
        self._transport_mock.read.return_value = expected_read
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected_exception)

    def test_open_excepts_with_wrong_data_length(self):
        """Test opening and closing the instrument"""
        # We expect as response MESSAGE_ID 0x0006 (_AptMsgHwGetInfo)
        expected_read = struct.pack("l", 0x0006) + b"\x00\x00"
        expected_exception = "Received incorrect message length for message id 0x{:04x} ".format(0x0006) +\
                             "(got {} bytes while expecting {} bytes)".format(len(expected_read), 90)
        # The data should include string "K10CR1" at right spot.
        self._transport_mock.read.return_value = expected_read
        with self.assertRaises(qmi.core.exceptions.QMI_InstrumentException) as exc:
            self.instr.open()

        self.assertEqual(str(exc.exception), expected_exception)


class TestThorlabsK10cr1Methods(unittest.TestCase):

    def setUp(self):
        qmi.start("TestK10cr1Context")
        self._transport_mock = unittest.mock.MagicMock(spec=QMI_SerialTransport)
        with unittest.mock.patch(
                'qmi.instruments.thorlabs.k10cr1.create_transport',
                return_value=self._transport_mock):
            self.instr: Thorlabs_K10Cr1 = qmi.make_instrument("instr", Thorlabs_K10Cr1, "transport_descriptor")
            self.instr = cast(Thorlabs_K10Cr1, self.instr)

        # We expect as response MESSAGE_ID 0x0006 (_AptMsgHwGetInfo) to 'open'
        expected_read = struct.pack("<l", 0x0006)
        # The request+data has to be 90 bytes long and should include string "K10CR1" at right spot.
        self._transport_mock.read.return_value = expected_read + b"\x00" * 2 + b"CR1\0K10" * 12
        self.instr.open()
        self._pack = "<l"
        self._empty = b"\x00" * 2
        self._write = b""
        # Clean up the mock for the tests.
        self._transport_mock.reset_mock()

    def tearDown(self):
        self.instr.close()
        qmi.stop()

    def test_get_idn(self):
        """Test the get_idn method."""
        # First two values are hardcoded, and the two latter are based on the standard return value for read in `setUp`
        expected_idn = ["Thorlabs", "K10CR1", 3232323, "49.82.67"]
        # We expect to write MESSAGE_ID 0x0005 (_AptMsgHwReqInfo)
        expected_write = struct.pack(self._pack, 0x0005) + self._write + b"P\x01"  # This is 5001 == 0x1389

        idn = self.instr.get_idn()
        self.assertEqual(idn.vendor, expected_idn[0])
        self.assertEqual(idn.model, expected_idn[1])
        self.assertEqual(idn.serial, expected_idn[2])
        self.assertEqual(idn.version, expected_idn[3])

        self._transport_mock.write.assert_called_with(expected_write)

    def test_get_motor_status(self):
        """Test the get_motor_status method. Test returning all statuses as 'True', and as 'False'"""
        # We expect to write MESSAGE_ID 0x0429 (_AptMsgReqStatusBits).
        # Bit 1 after x below is to identify as "Get" command.
        expected_write = struct.pack(self._pack, 0x10429) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x042A)
        # Let's get all expected values as "True". The bit order is in reverse pairs.
        self._transport_mock.read.return_value = expected_read + self._empty + b"\x00\x00\xfb\xff\x00\x81"

        motor_status = self.instr.get_motor_status()

        self.assertTrue(all(list(motor_status)))  # Check all statuses are "true"
        self._transport_mock.write.assert_called_with(expected_write)

        # Let's get all expected values as "False"
        self._transport_mock.read.reset_mock()
        self._transport_mock.read.return_value = expected_read + b"\x00\x00\x00\x00\x00\x00\x00\x00"

        motor_status = self.instr.get_motor_status()

        self.assertFalse(any(list(motor_status)))  # Check all statuses are "false"
        self._transport_mock.write.assert_called_with(expected_write)

    def test_identify(self):
        """Test the identify method"""
        # We expect to write MESSAGE_ID 0x0223 (_AptMsgIdentify).
        expected_write = struct.pack(self._pack, 0x10223) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x0223)
        # We have no other requirements from the read than the start being the same as the write
        self._transport_mock.read.return_value = expected_read + self._empty

        self.instr.identify()

        self._transport_mock.write.assert_called_with(expected_write)

    def test_get_chan_enable_state(self):
        """Test get_chan_enable_state method returns channel enabled state."""
        # _AptMsgReqChanEnableState 0x0211
        expected_write = struct.pack(self._pack, 0x10211) + self._write + b"P\x01"  # This is 5001 == 0x1389
        # _AptMsgGetChanEnableState 0x0212
        expected_read = struct.pack(self._pack, 0x0212)
        # We cannot change the enabled state so we content to having it as false
        self._transport_mock.read.return_value = expected_read + b"00"

        state = self.instr.get_chan_enable_state()

        self.assertFalse(state)
        self._transport_mock.write.assert_called_with(expected_write)

    def test_get_absolute_position(self):
        """Test get_absolute_position method returns expected position."""
        expected = 123.45
        # _AptMsgReqPosCounter 0x0411
        expected_write = struct.pack(self._pack, 0x10411) + self._write + b"P\x01"  # This is 5001 == 0x1389
        # _AptMsgGetPosCounter 0x0412
        expected_read = struct.pack(self._pack, 0x0412)
        # Add the position value at the end
        position_bs = b"\x00\x30\x01\x01"
        self._transport_mock.read.return_value = expected_read + b"\x00\x00\x00\x00" + position_bs

        position = self.instr.get_absolute_position()

        self.assertEqual(round(position, 2), expected)
        self._transport_mock.write.assert_called_with(expected_write)

    def test_get_velocity_params(self):
        """Test get_velocity_params returns velocity and acceleration."""
        expected_max_vel = 123.45
        expected_accel = 543.21
        # _AptMsgReqVelParams 0x0414
        expected_write = struct.pack(self._pack, 0x10414) + self._write + b"P\x01"  # This is 5001 == 0x1389
        # _AptMsgGetVelParams 0x0415
        expected_read = struct.pack(self._pack, 0x0415)
        # Manually define the bit strings
        max_vel = b"\x0a\xd3\xed\x35"
        accel = b"\x1d\x73\x0c\x00"
        self._transport_mock.read.return_value = expected_read + b"\x00\x00\x00\x00\x00\x00\x00\x00" + accel + max_vel

        velocity_params = self.instr.get_velocity_params()

        self.assertEqual(round(velocity_params.max_velocity, 2), expected_max_vel)
        self.assertEqual(round(velocity_params.acceleration, 2), expected_accel)
        self._transport_mock.write.assert_called_with(expected_write)

    def test_get_backlash_distance(self):
        """Test get_backlash_distance method returns expected value."""
        expected = 123.45
        # _AptMsgReqGenMoveParams 0x043B
        expected_write = struct.pack(self._pack, 0x1043B) + self._write + b"P\x01"  # This is 5001 == 0x1389
        # _AptMsgGetPosCounter 0x043C
        expected_read = struct.pack(self._pack, 0x043C)
        # We cannot change the enabled state so we content to having it as false
        self._transport_mock.read.return_value = expected_read + b"\x00\x00\x00\x00\x00\x30\x01\x01"

        distance = self.instr.get_backlash_distance()

        self.assertEqual(round(distance, 2), expected)
        self._transport_mock.write.assert_called_with(expected_write)

    def test_get_home_params(self):
        """Test get_home_params returns velocity and acceleration."""
        expected_home_dir = 1
        expected_limit_switch = 4
        expected_velocity = 123.45
        expected_offset = 543.21
        # _AptMsgReqHomeParams 0x0441
        expected_write = struct.pack(self._pack, 0x10441) + self._write + b"P\x01"  # This is 5001 == 0x1389
        # _AptMsgGetHomeParams 0x0442
        expected_read = struct.pack(self._pack, 0x0442) + b"\x00\x00\x00\x00"
        # manually define the bit strings
        home_dir = binascii.unhexlify(f"000{expected_home_dir}")
        limit_switch = binascii.unhexlify(f"000{expected_limit_switch}")
        home_vel = b"\x00\xd3\xed\x35"
        offset = b"\x00\xb0\x6b\x04"
        self._transport_mock.read.return_value = expected_read + home_dir[::-1] + limit_switch[::-1] + home_vel + offset

        home_params = self.instr.get_home_params()

        self._transport_mock.write.assert_called_with(expected_write)
        self.assertEqual(home_params.home_direction, expected_home_dir)
        self.assertEqual(home_params.limit_switch, expected_limit_switch)
        self.assertEqual(round(home_params.home_velocity, 2), expected_velocity)
        self.assertEqual(round(home_params.offset_distance, 2), expected_offset)

    def test_set_chan_enable_state(self):
        """Test set_chan_enable_state can be used to set state to True and False."""
        # We expect to write MESSAGE_ID 0x0210 (_AptMsgSetChanEnableState). Don't know why need to add 101 after x.
        expected_write = struct.pack(self._pack, 0x1010210) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x0210)
        # We have no other requirements from the read than the start being the same as the write
        self._transport_mock.read.return_value = expected_read + self._empty

        self.instr.set_chan_enable_state(True)

        self._transport_mock.write.assert_called_with(expected_write)

        # Let's try setting it as False as well. Note the different start now in 'expected_write'
        expected_write = struct.pack(self._pack, 0x2010210) + self._write + b"P\x01"  # This is 5001 == 0x1389

        self.instr.set_chan_enable_state(False)

        self._transport_mock.write.assert_called_with(expected_write)

    def test_set_velocity_params(self):
        """Test set_velocity_params can be used to set velocity and acceleration."""
        # Values to set
        velocity = 1.2345
        acceleration = 5.4321
        # We expect to write _AptMsgSetVelParams 0x0413. "e" after x below tells the data follows in 14 bits.
        data_def_bits = b"\xd0\x01\x01\x00"
        expected_read = struct.pack(self._pack, 0xe0413) + data_def_bits
        # velocity and acceleration get turned into ints in the command. Do the same here.
        vel_int = int(round(velocity * VELOCITY_FACTOR))
        acc_int = int(round(acceleration * ACCELERATION_FACTOR))
        # Then the command values are based on these ints. The length must be filled up to four bits.
        max_vel_hex = binascii.a2b_hex(hex(vel_int)[2:])
        accel_hex = binascii.a2b_hex(hex(acc_int)[2:])
        max_vel = b"\x00" * (4 - len(max_vel_hex)) + max_vel_hex
        accel = b"\x00" * (4 - len(accel_hex)) + accel_hex
        # Now we create the full expected write, with noting that the order is Little Endian (hence [::-1])
        expected_write = expected_read + b"\x00\x00\x00\x00" + accel[::-1] + max_vel[::-1]
        self._transport_mock.read.return_value = expected_read + self._empty
        # Act
        self.instr.set_velocity_params(velocity, acceleration)
        # Assert
        self._transport_mock.write.assert_called_with(expected_write)

    def test_set_velocity_params_values_out_of_range(self):
        """Test set_velocity_params raises exceptions at invalid values."""
        expected1 = "Invalid range for max_velocity"
        expected2 = "Invalid range for acceleration"

        velocity = 12.345
        acceleration = -43.210
        with self.assertRaises(ValueError) as exc1:
            self.instr.set_velocity_params(velocity, acceleration)

        valid_velocity = 1.2345
        with self.assertRaises(ValueError) as exc2:
            self.instr.set_velocity_params(valid_velocity, acceleration)

        self.assertEqual(str(exc1.exception), expected1)
        self.assertEqual(str(exc2.exception), expected2)

    def test_set_backlash_distance(self):
        """Test setting a backlash distance."""
        # Values to set
        distance = MICROSTEPS_PER_DEGREE / 20.0
        # We expect to write _AptMsgSetGenMoveParams 0x043A. "6" after x below tells the data follows in 6 bits.
        data_def_bits = b"\xd0\x01"
        expected_read = struct.pack(self._pack, 0x6043A) + data_def_bits
        # velocity and acceleration get turned into ints in the command. Do the same here.
        dist_int = int(round(distance * MICROSTEPS_PER_DEGREE))
        # Then the command values are based on these ints. The length must be filled up to four bits.
        dist_int_hex = binascii.a2b_hex((hex(dist_int))[2:])
        dist_bs = b"\x00" * (4 - len(dist_int_hex)) + dist_int_hex
        # Now we create the full expected write, with noting that the order is Little Endian (hence [::-1])
        expected_write = expected_read + b"\x01\x00" + dist_bs[::-1]
        self._transport_mock.read.return_value = expected_read
        # Act
        self.instr.set_backlash_distance(distance)
        # Assert
        self._transport_mock.write.assert_called_with(expected_write)

    def test_set_backlash_distance_out_of_range(self):
        """Test that the backlash distance excepts if not within uint32 range."""
        too_far_distances = [-2**31 / MICROSTEPS_PER_DEGREE, 2**31 / MICROSTEPS_PER_DEGREE]
        for distance in too_far_distances:
            with self.assertRaises(ValueError):
                self.instr.set_backlash_distance(distance)

    def test_set_home_params(self):
        """Test set_home_params can be used to set direction, limit switch, velocity and offset distance."""
        # Values to set
        home_dir = 2
        limit_switch = 1
        velocity = 1.2345
        offset_dist = 5.4321
        # We expect to write _AptMsgSetHomeParams 0x0440. "e" after x below tells the data follows in 14 bits.
        data_def_bits = b"\xd0\x01\x01\x00"
        expected_read = struct.pack(self._pack, 0xe0440) + data_def_bits
        # velocity and offset get turned into ints in the command. Do the same here.
        vel_int = int(round(velocity * VELOCITY_FACTOR))
        # Then the command values are based on these ints. The length must be filled up to four bits.
        home_vel_hex = binascii.a2b_hex(hex(vel_int)[2:])
        home_vel = b"\x00" * (4 - len(home_vel_hex)) + home_vel_hex
        offset = b"\x00\x0b\x51\x1f"
        home_dir_bs = binascii.unhexlify(f"000{home_dir}")
        limit_switch_bs = binascii.unhexlify(f"000{limit_switch}")
        # Now we create the full expected write, with noting that the order is Little Endian (hence [::-1])
        expected_write = expected_read + home_dir_bs[::-1] + limit_switch_bs[::-1] + home_vel[::-1] + offset[::-1]
        self._transport_mock.read.return_value = expected_read + self._empty
        # Act
        self.instr.set_home_params(home_dir, limit_switch, velocity, offset_dist)
        # Assert
        self._transport_mock.write.assert_called_with(expected_write)

    def test_set_home_params_excepts_with_values_out_of_range(self):
        """Test set_home_params excepts with invalid direction, limit switch, velocity and offset distance."""
        # Values to set
        home_dir_ok = 2
        limit_switch_ok = 1
        velocity_ok = 1.2345
        offset_dist_ok = 5.4321
        home_dir_nok = 0
        limit_switch_nok = 10
        home_velocity_noks = [0, 6]
        offset_noks = [-2**31 / MICROSTEPS_PER_DEGREE, 2**31 / MICROSTEPS_PER_DEGREE]
        # Test home_direction
        with self.assertRaises(ValueError):
            self.instr.set_home_params(home_dir_nok, limit_switch_ok, velocity_ok, offset_dist_ok)

        # Test limit_switch
        with self.assertRaises(ValueError):
            self.instr.set_home_params(home_dir_ok, limit_switch_nok, velocity_ok, offset_dist_ok)

        # Test home_velocity
        for home_vel in home_velocity_noks:
            with self.assertRaises(ValueError):
                self.instr.set_home_params(home_dir_ok, limit_switch_ok, home_vel, offset_dist_ok)

        # Test offset_distance
        for offset in offset_noks:
            with self.assertRaises(ValueError):
                self.instr.set_home_params(home_dir_ok, limit_switch_ok, velocity_ok, offset)

    def test_move_stop(self):
        """Test the move_stop method without immediate stop"""
        # We expect to write _AptMsgMoveStop 0x0465.
        expected_write = struct.pack(self._pack, 0x2010465) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x0465)
        # We have no other requirements from the read than the start being the same as the write
        self._transport_mock.read.return_value = expected_read + self._empty

        self.instr.move_stop()

        self._transport_mock.write.assert_called_with(expected_write)

    def test_move_stop_immediate(self):
        """Test the move_stop method with immediate stop"""
        # We expect to write _AptMsgMoveStop 0x0465.
        expected_write = struct.pack(self._pack, 0x1010465) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x0465)
        # We have no other requirements from the read than the start being the same as the write
        self._transport_mock.read.return_value = expected_read + self._empty

        self.instr.move_stop(True)

        self._transport_mock.write.assert_called_with(expected_write)

    def test_move_home(self):
        """Test the move_home method."""
        # We expect to write _AptMsgMoveHome 0x0443.
        expected_write = struct.pack(self._pack, 0x10443) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x0443)
        # We have no other requirements from the read than the start being the same as the write
        self._transport_mock.read.return_value = expected_read + self._empty

        self.instr.move_home()

        self._transport_mock.write.assert_called_with(expected_write)

    def test_move_relative(self):
        """Test moving a relative distance."""
        # Values to set
        distance = MICROSTEPS_PER_DEGREE / 20.0
        # We expect to write _AptMsgMoveRelative 0x0448. "6" after x below tells the data follows in 6 bits.
        data_def_bits = b"\xd0\x01"
        expected_read = struct.pack(self._pack, 0x60448) + data_def_bits
        # velocity and acceleration get turned into ints in the command. Do the same here.
        dist_int = int(round(distance * MICROSTEPS_PER_DEGREE))
        # Then the command values are based on these ints. The length must be filled up to four bits.
        dist_int_hex = binascii.a2b_hex((hex(dist_int))[2:])
        dist_bs = b"\x00" * (4 - len(dist_int_hex)) + dist_int_hex
        # Now we create the full expected write, with noting that the order is Little Endian (hence [::-1])
        expected_write = expected_read + b"\x01\x00" + dist_bs[::-1]
        self._transport_mock.read.return_value = expected_read
        # Act
        self.instr.move_relative(distance)
        # Assert
        self._transport_mock.write.assert_called_with(expected_write)

    def test_move_relative_out_of_range(self):
        """Test that the move_relative excepts if not within uint32 range."""
        too_far_distances = [-2**31 / MICROSTEPS_PER_DEGREE, 2**31 / MICROSTEPS_PER_DEGREE]
        for distance in too_far_distances:
            with self.assertRaises(ValueError):
                self.instr.move_relative(distance)

    def test_move_absolute(self):
        """Test moving a relative distance."""
        # Values to set
        distance = MICROSTEPS_PER_DEGREE / 20.0
        # We expect to write _AptMsgMoveAbsolute 0x0453. "6" after x below tells the data follows in 6 bits.
        data_def_bits = b"\xd0\x01"
        expected_read = struct.pack(self._pack, 0x60453) + data_def_bits
        # velocity and acceleration get turned into ints in the command. Do the same here.
        dist_int = int(round(distance * MICROSTEPS_PER_DEGREE))
        # Then the command values are based on these ints. The length must be filled up to four bits.
        dist_int_hex = binascii.a2b_hex((hex(dist_int))[2:])
        dist_bs = b"\x00" * (4 - len(dist_int_hex)) + dist_int_hex
        # Now we create the full expected write, with noting that the order is Little Endian (hence [::-1])
        expected_write = expected_read + b"\x01\x00" + dist_bs[::-1]
        self._transport_mock.read.return_value = expected_read
        # Act
        self.instr.move_absolute(distance)
        # Assert
        self._transport_mock.write.assert_called_with(expected_write)

    def test_move_absolute_out_of_range(self):
        """Test that the move_absolute excepts if not within uint32 range."""
        too_far_distances = [-2**31 / MICROSTEPS_PER_DEGREE, 2**31 / MICROSTEPS_PER_DEGREE]
        for distance in too_far_distances:
            with self.assertRaises(ValueError):
                self.instr.move_absolute(distance)

    def test_wait_move_complete(self):
        """See that a move is completed. We need to get False for all motor moves values."""
        # We expect to write MESSAGE_ID 0x0429 (_AptMsgReqStatusBits).
        # Bit 1 after x below is to identify as "Get" command.
        expected_write = struct.pack(self._pack, 0x10429) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x042A)
        # Let's get all expected values as "False"
        self._transport_mock.read.return_value = expected_read + b"\x00\x00\x00\x00\x00\x00\x00\x00"

        self.instr.wait_move_complete(0.0)

        self._transport_mock.write.assert_called_with(expected_write)

    def test_wait_move_complete_excepts_with_timeout(self):
        """See that a move is completed. We need to get True for all motor moves values."""
        expected_exception = "Timeout while waiting for end of move"
        # We expect to write MESSAGE_ID 0x0429 (_AptMsgReqStatusBits).
        # Bit 1 after x below is to identify as "Get" command.
        expected_write = struct.pack(self._pack, 0x10429) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x042A)
        # Let's get all expected values as "True". The bit order is in reverse pairs.
        self._transport_mock.read.return_value = expected_read + b"\x00\x00\x00\x00\xfb\xff\x00\x81"

        with self.assertRaises(qmi.core.exceptions.QMI_TimeoutException) as exc:
            self.instr.wait_move_complete(0.0)

        self._transport_mock.write.assert_called_with(expected_write)
        self.assertEqual(str(exc.exception), expected_exception)

    def test_wait_move_complete_excepts_with_timeout_2(self):
        """See that a move is completed. We need to suppress motor giving a response."""
        expected_read_calls = 2
        # We expect to write MESSAGE_ID 0x0429 (_AptMsgReqStatusBits).
        # Bit 1 after x below is to identify as "Get" command.
        expected_write = struct.pack(self._pack, 0x10429) + self._write + b"P\x01"  # This is 5001 == 0x1389
        expected_read = struct.pack(self._pack, 0x042A)
        unexpected_read = struct.pack(self._pack, 0x0429) + self._empty
        # Let's get all expected values as "False"
        self._transport_mock.read.side_effect = [unexpected_read, expected_read + b"\x00\x00\x00\x00\x00\x00\x00\x00"]

        self.instr.wait_move_complete(0.0)

        self._transport_mock.write.assert_called_with(expected_write)
        self.assertEqual(self._transport_mock.read.call_count, expected_read_calls)


if __name__ == '__main__':
    unittest.main()
